@@ -5,6 +5,7 @@ import asyncio
 import csv
 import glob
 import io
+import itertools
 import json
 import os
 import random
@@ -219,6 +220,15 @@ async def run_format():
             no_capture=True,
         )
 
+    c_files = files = list(
+        itertools.chain.from_iterable(
+            glob.glob(p, recursive=True)
+            for p in ["tools/scxtop/**/*.h", "tools/scxtop/**/*.c"]
+        )
+    )
+    if c_files:
+        await run_command(["clang-format", "-i"] + c_files, no_capture=True)
+
     await run_command(["git", "diff", "--exit-code"], no_capture=True)
     print("✓ Format completed successfully", flush=True)
 
@@ -227,7 +237,11 @@ async def run_build():
     """Build all targets."""
     print("Running build...", flush=True)
 
+    print("Building C schedulers...", flush=True)
+    await run_command(["make", "all"], no_capture=True)
+    print("Building Rust schedulers...", flush=True)
     await run_command(["cargo", "build", "--all-targets", "--locked"], no_capture=True)
+
     print("✓ Build completed successfully", flush=True)
 
 
@@ -247,8 +261,29 @@ async def run_clippy():
 
 async def run_tests():
     """Run the test suite."""
+    print("Checking that vmlinux.tar.zst is in sync...", flush=True)
+
+    with tempfile.TemporaryDirectory() as tempdir:
+        await run_command(
+            ["tar", "-xf", "rust/scx_utils/vmlinux.tar.zst", "-C", tempdir],
+            no_capture=True,
+        )
+        await run_command(
+            [
+                "diff",
+                "--no-dereference",
+                "-qr",
+                "scheds/vmlinux/",
+                f"{tempdir}/vmlinux/",
+            ],
+            no_capture=True,
+        )
+
     print("Running tests...", flush=True)
 
+    # Make sure the selftest is built in case the build was not already run.
+    await run_command(["cargo", "build", "-p", "scx_lib_selftests"], no_capture=True)
+
     await run_command(
         [
             "cargo",
@@ -261,7 +296,7 @@ async def run_tests():
     )
 
     # Get CPU count
-    cpu_count = min(os.cpu_count(), 16)
+    cpu_count = min(os.cpu_count() or 16, 16)
 
     await run_command_in_vm(
         "sched_ext/for-next",
@@ -270,7 +305,7 @@ async def run_tests():
             "test-in-vm",
         ],
         memory=10 * 1024 * 1024 * 1024,
-        cpus=min(os.cpu_count(), 16),
+        cpus=cpu_count,
         no_capture=True,
     )
 
@@ -739,7 +774,7 @@ async def main():
         else:
             await run_veristat()
     elif args.command == "test-in-vm":
-        run_tests_in_vm()
+        await run_tests_in_vm()
     elif args.command == "all":
         await run_all()
 
