@@ -3,6 +3,7 @@
  * Copyright (c) 2025 Meta Platforms, Inc. and affiliates.
  */
 #include <scx/common.bpf.h>
+#include <scx/bpf_arena_common.bpf.h>
 
 #include "selftest.h"
 
@@ -35,6 +36,12 @@ int arena_selftest(void)
 		return ret;
 	}
 
+	ret = scx_selftest_rbtree();
+	if (ret) {
+		bpf_printk("scx_selftest_rbtree failed with %d", ret);
+		return ret;
+	}
+
 	ret = scx_selftest_topology();
 	if (ret) {
 		bpf_printk("scx_selftest_topology failed with %d", ret);
@@ -45,3 +52,5 @@ int arena_selftest(void)
 
 	return 0;
 }
+
+char _license[] SEC("license") = "GPL";
