@@ -12,6 +12,10 @@ use anyhow::Context;
 use anyhow::Result;
 
 use std::ffi::c_ulong;
+use std::ffi::c_void;
+
+use std::os::fd::AsFd;
+use std::os::fd::AsRawFd;
 use std::sync::Arc;
 
 use scx_utils::init_libbpf_logging;
@@ -22,11 +26,16 @@ use scx_utils::NR_CPU_IDS;
 
 use simplelog::{ColorChoice, Config as SimplelogConfig, TermLogger, TerminalMode};
 
+use libbpf_rs::libbpf_sys;
+
 use libbpf_rs::skel::OpenSkel;
 use libbpf_rs::skel::SkelBuilder;
 use libbpf_rs::PrintLevel;
 use libbpf_rs::ProgramInput;
 
+const BPF_STDOUT: u32 = 1;
+const BPF_STDERR: u32 = 2;
+
 fn setup_arenas(skel: &mut BpfSkel<'_>) -> Result<()> {
     const STATIC_ALLOC_PAGES_GRANULARITY: c_ulong = 512;
     const TASK_SIZE: c_ulong = 42;
@@ -107,7 +116,7 @@ fn setup_topology_node(skel: &mut BpfSkel<'_>, mask: &[u64]) -> Result<()> {
     let output = skel.progs.arena_topology_node_init.test_run(input)?;
     if output.return_value != 0 {
         bail!(
-            "p2dq_topology_node_init returned {}",
+            "arena_topology_node_init returned {}",
             output.return_value as i32
         );
     }
@@ -152,6 +161,41 @@ fn setup_topology(skel: &mut BpfSkel<'_>) -> Result<()> {
     Ok(())
 }
 
+fn print_stream(skel: &mut BpfSkel<'_>, stream_id: u32) -> () {
+    let mut buf = vec![0u8; 4096];
+    let name = if stream_id == 1 { "OUTPUT" } else { "ERROR" };
+    let mut started = false;
+
+    loop {
+        let ret = unsafe {
+            libbpf_sys::bpf_prog_stream_read(
+                skel.progs.arena_selftest.as_fd().as_raw_fd(),
+                stream_id,
+                buf.as_mut_ptr() as *mut c_void,
+                buf.len() as u32,
+                std::ptr::null_mut(),
+            )
+        };
+        if ret < 0 {
+            eprintln!("STREAM {} UNAVAILABLE (REQUIRES >= v6.17)", name);
+            return;
+        }
+
+        if !started {
+            println!("===BEGIN STREAM {}===", name);
+            started = true;
+        }
+
+        if ret == 0 {
+            break;
+        }
+
+        print!("{}", String::from_utf8_lossy(&buf[..ret as usize]));
+    }
+
+    println!("\n====END STREAM  {}====", name);
+}
+
 fn main() {
     TermLogger::init(
         simplelog::LevelFilter::Info,
@@ -185,9 +229,14 @@ fn main() {
 
     let output = skel.progs.arena_selftest.test_run(input).unwrap();
     if output.return_value != 0 {
-        println!(
+        eprintln!(
             "Selftest returned {}, please check bpf tracelog for more details.",
             output.return_value as i32
         );
+    } else {
+        println!("Selftest successful.");
     }
+
+    print_stream(&mut skel, BPF_STDOUT);
+    print_stream(&mut skel, BPF_STDERR);
 }
