@@ -21,12 +21,14 @@ use std::sync::Once;
 
 use anyhow::Context;
 use anyhow::Result;
+use anyhow::bail;
 
 use plain::Plain;
 use procfs::process::all_processes;
 
 use libbpf_rs::OpenObject;
 use libbpf_rs::ProgramInput;
+use libbpf_rs::libbpf_sys::bpf_object_open_opts;
 
 use libc::{pthread_self, pthread_setschedparam, sched_param};
 
@@ -170,10 +172,6 @@ struct AlignedBuffer([u8; BUFSIZE]);
 
 static mut BUF: AlignedBuffer = AlignedBuffer([0; BUFSIZE]);
 
-// Special negative error code for libbpf to stop after consuming just one item from a BPF
-// ring buffer.
-const LIBBPF_STOP: i32 = -255;
-
 static SET_HANDLER: Once = Once::new();
 
 fn set_ctrlc_handler(shutdown: Arc<AtomicBool>) -> Result<(), anyhow::Error> {
@@ -190,10 +188,12 @@ fn set_ctrlc_handler(shutdown: Arc<AtomicBool>) -> Result<(), anyhow::Error> {
 impl<'cb> BpfScheduler<'cb> {
     pub fn init(
         open_object: &'cb mut MaybeUninit<OpenObject>,
+        open_opts: Option<bpf_object_open_opts>,
         exit_dump_len: u32,
         partial: bool,
         debug: bool,
         builtin_idle: bool,
+        name: &str,
     ) -> Result<Self> {
         let shutdown = Arc::new(AtomicBool::new(false));
         set_ctrlc_handler(shutdown.clone()).context("Error setting Ctrl-C handler")?;
@@ -201,7 +201,7 @@ impl<'cb> BpfScheduler<'cb> {
         // Open the BPF prog first for verification.
         let mut skel_builder = BpfSkelBuilder::default();
         skel_builder.obj_builder.debug(debug);
-        let mut skel = scx_ops_open!(skel_builder, open_object, rustland)?;
+        let mut skel = scx_ops_open!(skel_builder, open_object, rustland, open_opts)?;
 
         // Copy one item from the ring buffer.
         //
@@ -225,18 +225,8 @@ impl<'cb> BpfScheduler<'cb> {
                 BUF.0.copy_from_slice(data);
             }
 
-            // Return an unsupported error to stop early and consume only one item.
-            //
-            // NOTE: this is quite a hack. I wish libbpf would honor stopping after the first item
-            // is consumed, upon returning a non-zero positive value here, but it doesn't seem to
-            // be the case:
-            //
-            // https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/tools/lib/bpf/ringbuf.c?h=v6.8-rc5#n260
-            //
-            // Maybe we should fix this to stop processing items from the ring buffer also when a
-            // value > 0 is returned.
-            //
-            LIBBPF_STOP
+            // Return 0 to indicate successful completion of the copy.
+            0
         }
 
         // Check host topology to determine if we need to enable SMT capabilities.
@@ -255,6 +245,7 @@ impl<'cb> BpfScheduler<'cb> {
         skel.maps.rodata_data.as_mut().unwrap().khugepaged_pid = Self::khugepaged_pid();
         skel.maps.rodata_data.as_mut().unwrap().builtin_idle = builtin_idle;
         skel.maps.rodata_data.as_mut().unwrap().debug = debug;
+        let _ = Self::set_scx_ops_name(&mut skel.struct_ops.rustland_mut().name, name);
 
         // Attach BPF scheduler.
         let mut skel = scx_ops_load!(skel, rustland, uei)?;
@@ -302,6 +293,23 @@ impl<'cb> BpfScheduler<'cb> {
         })
     }
 
+    // Set the name of the scx ops.
+    fn set_scx_ops_name(dst: &mut [i8], s: &str) -> Result<()> {
+        if !s.is_ascii() {
+            bail!("name must be an ASCII string");
+        }
+
+        let bytes = s.as_bytes();
+        let n = bytes.len().min(dst.len().saturating_sub(1));
+
+        dst.fill(0);
+        for i in 0..n {
+            dst[i] = bytes[i] as i8;
+        }
+
+        Ok(())
+    }
+
     // Return the PID of khugepaged, if present, otherwise return 0.
     fn khugepaged_pid() -> u32 {
         let procs = match all_processes() {
@@ -431,61 +439,121 @@ impl<'cb> BpfScheduler<'cb> {
     // Counter of the online CPUs.
     #[allow(dead_code)]
     pub fn nr_online_cpus_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_online_cpus
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_online_cpus
     }
 
     // Counter of currently running tasks.
     #[allow(dead_code)]
     pub fn nr_running_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_running
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_running
     }
 
     // Counter of queued tasks.
     #[allow(dead_code)]
     pub fn nr_queued_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_queued
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_queued
     }
 
     // Counter of scheduled tasks.
     #[allow(dead_code)]
     pub fn nr_scheduled_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_scheduled
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_scheduled
     }
 
     // Counter of user dispatch events.
     #[allow(dead_code)]
     pub fn nr_user_dispatches_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_user_dispatches
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_user_dispatches
     }
 
     // Counter of user kernel events.
     #[allow(dead_code)]
     pub fn nr_kernel_dispatches_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_kernel_dispatches
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_kernel_dispatches
     }
 
     // Counter of cancel dispatch events.
     #[allow(dead_code)]
     pub fn nr_cancel_dispatches_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_cancel_dispatches
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_cancel_dispatches
     }
 
     // Counter of dispatches bounced to the shared DSQ.
     #[allow(dead_code)]
     pub fn nr_bounce_dispatches_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_bounce_dispatches
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_bounce_dispatches
     }
 
     // Counter of failed dispatch events.
     #[allow(dead_code)]
     pub fn nr_failed_dispatches_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_failed_dispatches
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_failed_dispatches
     }
 
     // Counter of scheduler congestion events.
     #[allow(dead_code)]
     pub fn nr_sched_congested_mut(&mut self) -> &mut u64 {
-        &mut self.skel.maps.bss_data.as_mut().unwrap().nr_sched_congested
+        &mut self
+            .skel
+            .maps
+            .bss_data
+            .as_mut()
+            .unwrap()
+            .nr_sched_congested
     }
 
     // Set scheduling class for the scheduler itself to SCHED_EXT
@@ -535,16 +603,24 @@ impl<'cb> BpfScheduler<'cb> {
     // Receive a task to be scheduled from the BPF dispatcher.
     #[allow(static_mut_refs)]
     pub fn dequeue_task(&mut self) -> Result<Option<QueuedTask>, i32> {
-        match self.queued.consume_raw() {
+        // Try to consume the first task from the ring buffer.
+        match self.queued.consume_raw_n(1) {
             0 => {
+                // Ring buffer is empty.
                 self.skel.maps.bss_data.as_mut().unwrap().nr_queued = 0;
                 Ok(None)
             }
-            LIBBPF_STOP => {
+            1 => {
                 // A valid task is received, convert data to a proper task struct.
                 let task = unsafe { EnqueuedMessage::from_bytes(&BUF.0).to_queued_task() };
-                self.skel.maps.bss_data.as_mut().unwrap().nr_queued =
-                    self.skel.maps.bss_data.as_ref().unwrap().nr_queued.saturating_sub(1);
+                self.skel.maps.bss_data.as_mut().unwrap().nr_queued = self
+                    .skel
+                    .maps
+                    .bss_data
+                    .as_ref()
+                    .unwrap()
+                    .nr_queued
+                    .saturating_sub(1);
 
                 Ok(Some(task))
             }
