@@ -59,6 +59,7 @@ pub use compat::ROOT_PREFIX;
 
 mod libbpf_logger;
 pub use libbpf_logger::init_libbpf_logging;
+pub mod libbpf_clap_opts;
 
 pub mod ravg;
 
@@ -92,7 +93,7 @@ pub mod mangoapp;
 pub mod misc;
 pub use misc::monitor_stats;
 pub use misc::normalize_load_metric;
-pub use misc::set_rlimit_infinity;
+pub use misc::try_set_rlimit_infinity;
 
 mod netdev;
 pub use netdev::read_netdevs;
