@@ -1,7 +1,7 @@
 use anyhow::Result;
-use anyhow::{anyhow, bail};
-use libc;
+use anyhow::{anyhow, Context};
 use log::{info, warn};
+use nix::sys::resource::{getrlimit, setrlimit, Resource, RLIM_INFINITY};
 use scx_stats::prelude::*;
 use serde::Deserialize;
 use std::path::Path;
@@ -44,18 +44,15 @@ where
         while !should_exit() {
             let stats = match client.request::<T>("stats", stats_args.to_owned()) {
                 Ok(v) => v,
-                Err(e) => match e.downcast_ref::<std::io::Error>() {
-                    Some(ioe) => {
+                Err(e) => {
+                    if let Some(ioe) = e.downcast_ref::<std::io::Error>() {
                         info!("Connection to stats_server failed ({ioe})");
-                        sleep(Duration::from_secs(1));
-                        break;
+                    } else {
+                        warn!("Error handling stats_server result: {e}");
                     }
-                    None => {
-                        warn!("error on handling stats_server result {e}");
-                        sleep(Duration::from_secs(1));
-                        break;
-                    }
-                },
+                    sleep(Duration::from_secs(1));
+                    break;
+                }
             };
             output(stats)?;
             sleep(intv);
@@ -65,18 +62,18 @@ where
     Ok(())
 }
 
-pub fn set_rlimit_infinity() {
-    unsafe {
-        // Call setrlimit to set the locked-in-memory limit to unlimited.
-        let new_rlimit = libc::rlimit {
-            rlim_cur: libc::RLIM_INFINITY,
-            rlim_max: libc::RLIM_INFINITY,
-        };
-        let res = libc::setrlimit(libc::RLIMIT_MEMLOCK, &new_rlimit);
-        if res != 0 {
-            panic!("setrlimit failed with error code: {res}");
+pub fn try_set_rlimit_infinity() {
+    // Increase MEMLOCK size since the BPF scheduler might use
+    // more than the current limit
+    if setrlimit(Resource::RLIMIT_MEMLOCK, RLIM_INFINITY, RLIM_INFINITY).is_err() {
+        // If there is an error in expanding rlimit to infinity,
+        // show the current rlimits then proceed.
+        if let Ok((soft, hard)) = getrlimit(Resource::RLIMIT_MEMLOCK) {
+            warn!("Current MEMLOCK limit: soft={soft}, hard={hard}");
+        } else {
+            warn!("Cannot change or query MEMLOCK limit");
         }
-    };
+    }
 }
 
 /// Read a file and parse its content into the specified type.
@@ -90,20 +87,13 @@ where
     T: std::str::FromStr,
     T::Err: std::error::Error + Send + Sync + 'static,
 {
-    let val = match std::fs::read_to_string(path) {
-        Ok(val) => val,
-        Err(_) => {
-            bail!("Failed to open or read file {:?}", path);
-        }
-    };
-    let val = val.trim_end_matches('\0');
+    let val = std::fs::read_to_string(path)
+        .with_context(|| format!("Failed to open or read file: {}", path.display()))?;
 
-    match val.trim().parse::<T>() {
-        Ok(parsed) => Ok(parsed),
-        Err(_) => {
-            bail!("Failed to parse content '{}' from {:?}", val.trim(), path);
-        }
-    }
+    let val = val.trim_end_matches('\0').trim();
+
+    val.parse::<T>()
+        .with_context(|| format!("Failed to parse content '{}' from {}", val, path.display()))
 }
 
 pub fn read_file_usize_vec(path: &Path, separator: char) -> Result<Vec<usize>> {
