@@ -82,7 +82,6 @@ use log::warn;
 use sscanf::sscanf;
 use std::collections::BTreeMap;
 use std::path::Path;
-use std::path::PathBuf;
 use std::sync::Arc;
 
 #[cfg(feature = "gpu-topology")]
@@ -370,7 +369,7 @@ fn cpus_online() -> Result<Cpumask> {
     Cpumask::from_cpulist(&online)
 }
 
-fn get_cache_id(topo_ctx: &mut TopoCtx, cache_level_path: &PathBuf, cache_level: usize) -> usize {
+fn get_cache_id(topo_ctx: &mut TopoCtx, cache_level_path: &Path, cache_level: usize) -> usize {
     // Check if the cache id is already cached
     let id_map = match cache_level {
         2 => &mut topo_ctx.l2_ids,
@@ -404,7 +403,7 @@ fn get_cache_id(topo_ctx: &mut TopoCtx, cache_level_path: &PathBuf, cache_level:
     id
 }
 
-fn get_per_cpu_cache_size(cache_path: &PathBuf) -> Result<usize> {
+fn get_per_cpu_cache_size(cache_path: &Path) -> Result<usize> {
     let path_str = cache_path.to_str().unwrap();
     let paths = glob(&(path_str.to_owned() + "/index[0-9]*"))?;
     let mut tot_size = 0;
@@ -526,7 +525,7 @@ fn create_insert_cpu(
         llc_id: *llc_id,
         node_id: node.id,
         kernel_id: core_kernel_id,
-        cluster_id: cluster_id,
+        cluster_id,
     }));
     let core_mut = Arc::get_mut(core).unwrap();
 
@@ -572,7 +571,7 @@ fn read_cpu_ids() -> Result<Vec<usize>> {
     let cpu_paths = glob(&path)?;
     for cpu_path in cpu_paths.filter_map(Result::ok) {
         let cpu_str = cpu_path.to_str().unwrap().trim();
-        if *ROOT_PREFIX == "" {
+        if ROOT_PREFIX.is_empty() {
             match sscanf!(cpu_str, "/sys/devices/system/cpu/cpu{usize}") {
                 Ok(val) => cpu_ids.push(val),
                 Err(_) => {
@@ -750,7 +749,7 @@ fn create_numa_nodes(
     let numa_paths = glob(&path)?;
     for numa_path in numa_paths.filter_map(Result::ok) {
         let numa_str = numa_path.to_str().unwrap().trim();
-        let node_id = if *ROOT_PREFIX == "" {
+        let node_id = if ROOT_PREFIX.is_empty() {
             match sscanf!(numa_str, "/sys/devices/system/node/node{usize}") {
                 Ok(val) => val,
                 Err(_) => {
@@ -801,7 +800,7 @@ fn create_numa_nodes(
         let mut cpu_ids = vec![];
         for cpu_path in cpu_paths.filter_map(Result::ok) {
             let cpu_str = cpu_path.to_str().unwrap().trim();
-            let cpu_id = if *ROOT_PREFIX == "" {
+            let cpu_id = if ROOT_PREFIX.is_empty() {
                 match sscanf!(cpu_str, "/sys/devices/system/node/node{usize}/cpu{usize}") {
                     Ok((_, val)) => val,
                     Err(_) => {
