@@ -14,6 +14,8 @@ enum chaos_consts {
 	CHAOS_DSQ_BASE		= 1 << CHAOS_DSQ_BASE_SHIFT,
 
 	CHAOS_NUM_PPIDS_CHECK	= 1 << 20,
+
+	CHAOS_MAX_RAND_ATTEMPTS	= 512,
 };
 
 enum chaos_match {
@@ -30,6 +32,7 @@ enum chaos_trait_kind {
 	CHAOS_TRAIT_RANDOM_DELAYS,
 	CHAOS_TRAIT_CPU_FREQ,
 	CHAOS_TRAIT_DEGRADATION,
+	CHAOS_TRAIT_KPROBE_RANDOM_DELAYS,
 	CHAOS_TRAIT_MAX,
 };
 
@@ -49,6 +52,7 @@ enum chaos_stat_idx {
 	CHAOS_STAT_TRAIT_DEGRADATION,
 	CHAOS_STAT_CHAOS_EXCLUDED,
 	CHAOS_STAT_CHAOS_SKIPPED,
+	CHAOS_STAT_KPROBE_RANDOM_DELAYS,
 	CHAOS_STAT_TIMER_KICKS,
 	CHAOS_NR_STATS,
 };
