@@ -51,6 +51,8 @@ const volatile u32 degradation_freq_frac32 = 1;
 const volatile u64 degradation_frac7 = 0;
 
 const volatile u32 kprobe_delays_freq_frac32 = 1;
+const volatile u64 kprobe_delays_min_ns = 1;
+const volatile u64 kprobe_delays_max_ns = 2;
 
 #define MIN(x, y) ((x) < (y) ? (x) : (y))
 #define MAX(x, y) ((x) > (y) ? (x) : (y))
@@ -86,6 +88,46 @@ struct chaos_task_ctx *lookup_create_chaos_task_ctx(struct task_struct *p)
 	return bpf_task_storage_get(&chaos_task_ctxs, p, NULL, BPF_LOCAL_STORAGE_GET_F_CREATE);
 }
 
+static __always_inline u64 chaos_get_prandom_u64_limit(u64 s)
+{
+	// Implementation of Lemire's algorithm 5 without 128-bit arithmetic.
+	// See https://arxiv.org/pdf/1805.10941v2 for details.
+	// Uses a bounded loop given this is BPF, but given the loop should
+	// rarely be entered this is fine.
+	u64 x, m_low, m_high;
+	u64 t;
+
+	x = get_prandom_u64();
+
+	// Compute 64-bit multiplication high and low parts
+	// m = x * s, split into m_high and m_low
+	m_high = ((x >> 32) * (s >> 32)) +
+		 (((x & 0xFFFFFFFF) * (s >> 32)) >> 32) +
+		 (((x >> 32) * (s & 0xFFFFFFFF)) >> 32);
+	m_low = x * s;
+
+	if (m_low < s) {
+		t = ((u64)(-s)) % s;
+		bpf_repeat(CHAOS_MAX_RAND_ATTEMPTS) {
+			if (m_low >= t)
+				break;
+
+			x = get_prandom_u64();
+			m_high = ((x >> 32) * (s >> 32)) +
+				 (((x & 0xFFFFFFFF) * (s >> 32)) >> 32) +
+				 (((x >> 32) * (s & 0xFFFFFFFF)) >> 32);
+			m_low = x * s;
+		}
+	}
+
+	return m_high;
+}
+
+static __always_inline u64 chaos_get_uniform_u64(u64 min, u64 max)
+{
+	return min + chaos_get_prandom_u64_limit(max - min + 1);
+}
+
 static __always_inline void chaos_stat_inc(enum chaos_stat_idx stat)
 {
 	u64 *cnt_p = bpf_map_lookup_elem(&chaos_stats, &stat);
@@ -114,7 +156,8 @@ static __always_inline enum chaos_trait_kind choose_chaos(struct chaos_task_ctx
 
 static __always_inline bool chaos_trait_skips_select_cpu(struct chaos_task_ctx *taskc)
 {
-	return taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS;
+	return taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS || 
+		taskc->next_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS;
 }
 
 static __always_inline u64 get_cpu_delay_dsq(int cpu_idx)
@@ -251,18 +294,10 @@ out:
 }
 
 __weak s32 enqueue_random_delay(struct task_struct *p __arg_trusted, u64 enq_flags,
-				struct chaos_task_ctx *taskc __arg_nonnull)
+				struct chaos_task_ctx *taskc __arg_nonnull, u64 min_ns, u64 max_ns)
 {
-	u64 rand64 = ((u64)bpf_get_prandom_u32() << 32) | bpf_get_prandom_u32();
-
-	u64 vtime = bpf_ktime_get_ns() + random_delays_min_ns;
-	if (random_delays_min_ns != random_delays_max_ns) {
-		vtime += rand64 % (random_delays_max_ns - random_delays_min_ns);
-	}
-
+	u64 vtime = chaos_get_uniform_u64(min_ns, max_ns);
 	scx_bpf_dsq_insert_vtime(p, get_cpu_delay_dsq(-1), 0, vtime, enq_flags);
-	chaos_stat_inc(CHAOS_STAT_TRAIT_RANDOM_DELAYS);
-
 	return true;
 }
 
@@ -272,10 +307,22 @@ __weak s32 enqueue_chaotic(struct task_struct *p __arg_trusted, u64 enq_flags,
 	bool out;
 
 	switch (taskc->next_trait) {
+	case CHAOS_TRAIT_KPROBE_RANDOM_DELAYS:
+		out = enqueue_random_delay(p,
+					   enq_flags,
+					   taskc,
+					   kprobe_delays_min_ns,
+					   kprobe_delays_max_ns);
+		chaos_stat_inc(CHAOS_STAT_KPROBE_RANDOM_DELAYS);
+		break;
 	case CHAOS_TRAIT_RANDOM_DELAYS:
-		out = enqueue_random_delay(p, enq_flags, taskc);
+		out = enqueue_random_delay(p,
+					   enq_flags,
+					   taskc,
+					   random_delays_min_ns,
+					   random_delays_max_ns);
+		chaos_stat_inc(CHAOS_STAT_TRAIT_RANDOM_DELAYS);
 		break;
-
 	case CHAOS_TRAIT_NONE:
 		chaos_stat_inc(CHAOS_STAT_CHAOS_SKIPPED);
 		out = false;
@@ -339,7 +386,7 @@ static int chaos_timer_check_queues_callback(void *map, int key, struct timer_wr
 	u64 this_next_trigger_time;
 	int cpu_idx;
 
-	bpf_for(cpu_idx, 0, nr_cpus) {
+	bpf_for(cpu_idx, 0, topo_config.nr_cpus) {
 		this_next_trigger_time = check_dsq_times(cpu_idx);
 		next_trigger_time = MAX(next_trigger_time, this_next_trigger_time);
 	}
@@ -364,7 +411,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init)
 	struct cpu_ctx *cpuc;
 	int timer_id, ret, i;
 
-	bpf_for(i, 0, nr_cpus) {
+	bpf_for(i, 0, topo_config.nr_cpus) {
 		if (!(cpuc = lookup_cpu_ctx(i)) ||
 		    !(llcx = lookup_llc_ctx(cpuc->llc_id)))
 			return -EINVAL;
@@ -396,29 +443,53 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(chaos_init)
 	return p2dq_init_impl();
 }
 
-static __always_inline __maybe_unused void complete_p2dq_enqueue_move(struct enqueue_promise *pro,
+/* 
+ * Cleanup any outstanding work left over by the p2dq promise when we do not
+ * plan to complete it.
+ */
+static __always_inline void destroy_p2dq_enqueue_promise(struct enqueue_promise *pro)
+{
+	// If the idle bit of a CPU has already been cleared but we don't plan
+	// to execute a task on it we should kick the CPU. If the CPU goes to
+	// sleep again it will reset the kernel managed idle state.
+	if (pro->has_cleared_idle)
+		scx_bpf_kick_cpu(pro->cpu, SCX_KICK_IDLE);
+}
+
+static __always_inline void complete_p2dq_enqueue_move(struct enqueue_promise *pro,
 						       struct bpf_iter_scx_dsq *it__iter,
 						       struct task_struct *p)
 {
 	switch (pro->kind) {
 	case P2DQ_ENQUEUE_PROMISE_COMPLETE:
-		goto out;
+		scx_bpf_error("chaos: delayed async_p2dq_enqueue returned COMPLETE"
+			      " after a task was placed in the delay dsq!");
+		break;
 	case P2DQ_ENQUEUE_PROMISE_FIFO:
 		__COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter, *MEMBER_VPTR(pro->fifo, .slice_ns));
 		__COMPAT_chaos_scx_bpf_dsq_move(it__iter, p, pro->fifo.dsq_id, pro->fifo.enq_flags);
-		goto out;
+		break;
 	case P2DQ_ENQUEUE_PROMISE_VTIME:
 		__COMPAT_chaos_scx_bpf_dsq_move_set_slice(it__iter, pro->vtime.slice_ns);
 		__COMPAT_chaos_scx_bpf_dsq_move_set_vtime(it__iter, pro->vtime.vtime);
 		__COMPAT_chaos_scx_bpf_dsq_move_vtime(it__iter, p, pro->vtime.dsq_id, pro->vtime.enq_flags);
-		goto out;
+		break;
+	case P2DQ_ENQUEUE_PROMISE_ATQ_FIFO:
+	case P2DQ_ENQUEUE_PROMISE_ATQ_VTIME:
+		scx_bpf_error("chaos: ATQs not supported");
+		break;
+	case P2DQ_ENQUEUE_PROMISE_FAILED:
+		scx_bpf_error("chaos: delayed async_p2dq_enqueue failed");
+		break;
 	}
 
-out:
+	if (pro->kick_idle)
+		scx_bpf_kick_cpu(pro->cpu, SCX_KICK_IDLE);
+
 	pro->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
 }
 
-__weak __maybe_unused int async_p2dq_enqueue_weak(struct enqueue_promise *ret __arg_nonnull,
+__weak int async_p2dq_enqueue_weak(struct enqueue_promise *ret __arg_nonnull,
 				   struct task_struct *p __arg_trusted,
 				   u64 enq_flags)
 {
@@ -428,12 +499,12 @@ __weak __maybe_unused int async_p2dq_enqueue_weak(struct enqueue_promise *ret __
 
 void BPF_STRUCT_OPS(chaos_dispatch, s32 cpu, struct task_struct *prev)
 {
+	struct enqueue_promise promise;
 	struct chaos_task_ctx *taskc;
 	struct task_struct *p;
 	u64 now = bpf_ktime_get_ns();
-	u64 initial_dsq = get_cpu_delay_dsq(-1);
 
-	bpf_for_each(scx_dsq, p, initial_dsq, 0) {
+	bpf_for_each(scx_dsq, p, get_cpu_delay_dsq(-1), 0) {
 		p = bpf_task_from_pid(p->pid);
 		if (!p)
 			continue;
@@ -452,7 +523,8 @@ void BPF_STRUCT_OPS(chaos_dispatch, s32 cpu, struct task_struct *prev)
 		// restore vtime to p2dq's timeline
 		p->scx.dsq_vtime = taskc->p2dq_vtime;
 
-		scx_bpf_dsq_move_to_local(initial_dsq);
+		async_p2dq_enqueue_weak(&promise, p, taskc->enq_flags);
+		complete_p2dq_enqueue_move(&promise, BPF_FOR_EACH_ITER, p);
 		bpf_task_release(p);
 	}
 
@@ -475,10 +547,13 @@ void BPF_STRUCT_OPS(chaos_enqueue, struct task_struct *p __arg_trusted, u64 enq_
 	async_p2dq_enqueue(&promise, p, enq_flags);
 	if (promise.kind == P2DQ_ENQUEUE_PROMISE_COMPLETE)
 		return;
+	if (promise.kind == P2DQ_ENQUEUE_PROMISE_FAILED)
+		goto cleanup;
 
-	if (taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS &&
+	if ((taskc->next_trait == CHAOS_TRAIT_RANDOM_DELAYS ||
+		taskc->next_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS) &&
 	    enqueue_chaotic(p, enq_flags, taskc))
-		return;
+		goto cleanup;
 
 	// NOTE: this may not work for affinitized tasks because p2dq does
 	// direct dispatch in some situations.
@@ -498,6 +573,10 @@ void BPF_STRUCT_OPS(chaos_enqueue, struct task_struct *p __arg_trusted, u64 enq_
 	}
 
 	complete_p2dq_enqueue(&promise, p);
+	return;
+
+cleanup:
+	destroy_p2dq_enqueue_promise(&promise);
 }
 
 void BPF_STRUCT_OPS(chaos_runnable, struct task_struct *p, u64 enq_flags)
@@ -561,7 +640,7 @@ void BPF_STRUCT_OPS(chaos_tick, struct task_struct *p)
 	if (!(taskc = lookup_create_chaos_task_ctx(p)))
 		return;
 
-	if (taskc->pending_trait == CHAOS_TRAIT_RANDOM_DELAYS)
+	if (taskc->pending_trait == CHAOS_TRAIT_KPROBE_RANDOM_DELAYS)
 		p->scx.slice = 0;
 }
 
@@ -588,6 +667,8 @@ SCX_OPS_DEFINE(chaos,
 	       .select_cpu		= (void *)chaos_select_cpu,
 	       .tick 		    	= (void *)chaos_tick,
 
+	       .update_idle		= (void *)p2dq_update_idle,
+	       .cpu_release		= (void *)p2dq_cpu_release,
 	       .exit_task		= (void *)p2dq_exit_task,
 	       .exit			= (void *)p2dq_exit,
 	       .running			= (void *)chaos_running,
@@ -612,7 +693,7 @@ int generic(struct pt_regs *ctx)
 
 	u32 roll = bpf_get_prandom_u32();
 	if (roll <= kprobe_delays_freq_frac32) {
-		taskc->pending_trait = CHAOS_TRAIT_RANDOM_DELAYS;
+		taskc->pending_trait = CHAOS_TRAIT_KPROBE_RANDOM_DELAYS;
 		dbg("GENERIC: setting pending_trait to RANDOM_DELAYS - task[%d]", p->pid);
 	}
 
