@@ -32,10 +32,11 @@ use log::{debug, info, warn};
 use scx_stats::prelude::*;
 use scx_utils::build_id;
 use scx_utils::compat;
+use scx_utils::libbpf_clap_opts::LibbpfOpts;
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
-use scx_utils::set_rlimit_infinity;
+use scx_utils::try_set_rlimit_infinity;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
 use scx_utils::CoreType;
@@ -67,10 +68,33 @@ struct Opts {
     slice_lag_us: u64,
 
     /// CPU busy threshold.
+    ///
+    /// Specifies the CPU utilization percentage (0-100%) at which the scheduler considers the
+    /// system to be busy.
+    ///
+    /// When the average CPU utilization reaches this threshold, the scheduler switches from using
+    /// multiple per-CPU round-robin dispatch queues (which favor locality and reduced locking
+    /// contention) to a global deadline-based dispatch queue (which improves load balancing).
+    ///
+    /// The global dispatch queue can increase task migrations and improve responsiveness for
+    /// interactive tasks under heavy load. Lower values make the scheduler switch to deadline
+    /// mode sooner, improving overall responsiveness at the cost of reducing single-task
+    /// performance due to the additional migrations. Higher values makes task more "sticky" to
+    /// their CPU, improving workloads that benefit from cache locality.
+    ///
+    /// A higher value is recommended for server-type workloads, while a lower value is recommended
+    /// for interactive-type workloads.
     #[clap(short = 'c', long, default_value = "75")]
     cpu_busy_thresh: u64,
 
-    /// Polling time (ms). Value is clamped to the range [10 .. 1000].
+    /// Polling time (ms) to refresh the CPU utilization.
+    ///
+    /// This interval determines how often the scheduler refreshes the CPU utilization that is
+    /// compared with the CPU busy threshold (option -c) to decide if the system is busy or not
+    /// and trigger the switch between using multiple per-CPU dispatch queues or a single global
+    /// deadline-based dispatch queue.
+    ///
+    /// Value is clamped to the range [10 .. 1000].
     ///
     /// 0 = disabled.
     #[clap(short = 'p', long, default_value = "250")]
@@ -107,7 +131,7 @@ struct Opts {
 
     /// Enable preferred idle CPU scanning.
     ///
-    /// With this opition enabled, the scheduler will prioritize assigning tasks to higher-ranked
+    /// With this option enabled, the scheduler will prioritize assigning tasks to higher-ranked
     /// cores before considering lower-ranked ones.
     #[clap(short = 'P', long, action = clap::ArgAction::SetTrue)]
     preferred_idle_scan: bool,
@@ -164,6 +188,9 @@ struct Opts {
     /// Show descriptions for statistics.
     #[clap(long)]
     help_stats: bool,
+
+    #[clap(flatten, next_help_heading = "Libbpf Options")]
+    pub libbpf: LibbpfOpts,
 }
 
 #[derive(PartialEq)]
@@ -286,7 +313,7 @@ struct Scheduler<'a> {
 
 impl<'a> Scheduler<'a> {
     fn init(opts: &'a Opts, open_object: &'a mut MaybeUninit<OpenObject>) -> Result<Self> {
-        set_rlimit_infinity();
+        try_set_rlimit_infinity();
 
         // Initialize CPU topology.
         let topo = Topology::new().unwrap();
@@ -310,7 +337,8 @@ impl<'a> Scheduler<'a> {
         // Initialize BPF connector.
         let mut skel_builder = BpfSkelBuilder::default();
         skel_builder.obj_builder.debug(opts.verbose);
-        let mut skel = scx_ops_open!(skel_builder, open_object, cosmos_ops)?;
+        let open_opts = opts.libbpf.clone().into_bpf_open_opts();
+        let mut skel = scx_ops_open!(skel_builder, open_object, cosmos_ops, open_opts)?;
 
         skel.struct_ops.cosmos_ops_mut().exit_dump_len = opts.exit_dump_len;
 
@@ -517,7 +545,7 @@ impl<'a> Scheduler<'a> {
 
 impl Drop for Scheduler<'_> {
     fn drop(&mut self) {
-        info!("Unregister {} scheduler", SCHEDULER_NAME);
+        info!("Unregister {SCHEDULER_NAME} scheduler");
     }
 }
 
