@@ -1,6 +1,6 @@
 /* SPDX-License-Identifier: GPL-2.0 */
 /*
- * Copyright (c) 2024 Andrea Righi <arighi@nvidia.com>
+ * Copyright (c) 2024 Andrea Righi <andrea.righi@linux.dev>
  *
  * This software may be used and distributed according to the terms of the GNU
  * General Public License version 2.
@@ -35,7 +35,12 @@ typedef signed long s64;
 typedef int pid_t;
 #endif /* __VMLINUX_H__ */
 
+struct cpu_arg {
+	s32 cpu_id;
+};
+
 struct domain_arg {
+	s32 lvl_id;
 	s32 cpu_id;
 	s32 sibling_cpu_id;
 };
