@@ -9,6 +9,16 @@
 
 #define DSQ_FLAG_NODE	(1LLU << 32)
 
+extern unsigned CONFIG_HZ __kconfig;
+
+/*
+ * Return the time interval between two ticks in ns.
+ */
+static inline u64 tick_interval_ns(void)
+{
+	return NSEC_PER_SEC / CONFIG_HZ;
+}
+
 /*
  * Thresholds for applying hysteresis to CPU performance scaling:
  *  - CPUFREQ_LOW_THRESH: below this level, reduce performance to minimum
@@ -240,6 +250,20 @@ struct {
 	__type(value, struct throttle_timer);
 } throttle_timer SEC(".maps");
 
+/*
+ * Timer used to preempt CPUs in tickless mode.
+ */
+struct tickless_timer {
+	struct bpf_timer timer;
+};
+
+struct {
+	__uint(type, BPF_MAP_TYPE_ARRAY);
+	__uint(max_entries, 1);
+	__type(key, u32);
+	__type(value, struct tickless_timer);
+} tickless_timer SEC(".maps");
+
 /*
  * Per-node context.
  */
@@ -485,7 +509,8 @@ static inline u64 nr_tasks_waiting(s32 cpu)
 {
 	int node = __COMPAT_scx_bpf_cpu_node(cpu);
 
-	return scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu)) +
+	return scx_bpf_dsq_nr_queued(SCX_DSQ_LOCAL_ON | cpu) +
+	       scx_bpf_dsq_nr_queued(cpu_to_dsq(cpu)) +
 	       scx_bpf_dsq_nr_queued(node_to_dsq(node));
 }
 
@@ -493,14 +518,22 @@ static inline u64 nr_tasks_waiting(s32 cpu)
  * Return the time slice that can be assigned to a task queued to @dsq_id
  * DSQ.
  */
-static inline u64 task_slice(s32 cpu)
+static inline u64 task_slice(const struct task_struct *p, s32 cpu)
 {
 	u64 nr_wait = nr_tasks_waiting(cpu);
 
-	if (!nr_wait)
-		return tickless_sched ? SCX_SLICE_INF : slice_max;
+	/*
+	 * Scale by task weight only when -n / native_priority is set.
+	 * Since we don't have preemption, if there's a big gap between
+	 * priorities certain tasks may get a really long time slice and
+	 * this can hurt interactivity.
+	 */
+	u64 smax = native_priority ? scale_by_task_weight(p, slice_max) : slice_max;
 
-	return MAX(slice_max / nr_wait, slice_min);
+	if (!nr_wait)
+		return tickless_sched ? SCX_SLICE_INF : smax;
+
+	return MAX(smax / nr_wait, slice_min);
 }
 
 /*
@@ -1095,7 +1128,7 @@ s32 BPF_STRUCT_OPS(flash_select_cpu, struct task_struct *p,
 
 	cpu = pick_idle_cpu(p, tctx, prev_cpu, wake_flags, &is_idle);
 	if (rr_sched || (is_idle && can_direct_dispatch(cpu))) {
-		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, task_slice(cpu), 0);
+		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, task_slice(p, cpu), 0);
 		__sync_fetch_and_add(&nr_direct_dispatches, 1);
 	}
 
@@ -1165,6 +1198,14 @@ static bool kick_idle_cpu(const struct task_struct *p, const struct task_ctx *tc
 static bool is_cpu_busy(s32 cpu)
 {
 	const struct cpu_ctx *cctx;
+
+	/*
+	 * If no other tasks are contending for this CPU, consider the CPU
+	 * as non-busy.
+	 */
+	if (!nr_tasks_waiting(cpu))
+		return false;
+
 	/*
 	 * Determine whether a CPU is considered busy using the following logic:
 	 *  - if a fixed threshold is provided (@cpu_busy_thresh), use it
@@ -1260,7 +1301,7 @@ static bool try_direct_dispatch(struct task_struct *p, struct task_ctx *tctx,
 	 * If the CPU is not busy dispatch tasks in a round-robin fashion.
 	 */
 	if (direct_dispatch && !is_cpu_busy(cpu)) {
-		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, task_slice(cpu), enq_flags);
+		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, task_slice(p, cpu), enq_flags);
 		__sync_fetch_and_add(&nr_direct_dispatches, 1);
 		dispatched = true;
 
@@ -1273,7 +1314,7 @@ static bool try_direct_dispatch(struct task_struct *p, struct task_ctx *tctx,
 	 */
 	if (is_pcpu_task(p)) {
 		if (can_direct_dispatch(cpu) && scx_bpf_test_and_clear_cpu_idle(cpu)) {
-			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, task_slice(cpu), enq_flags);
+			scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, task_slice(p, cpu), enq_flags);
 			__sync_fetch_and_add(&nr_direct_dispatches, 1);
 			dispatched = true;
 
@@ -1295,7 +1336,7 @@ static bool try_direct_dispatch(struct task_struct *p, struct task_ctx *tctx,
 		return false;
 
 	if (can_direct_dispatch(cpu)) {
-		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, task_slice(cpu), 0);
+		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, task_slice(p, cpu), 0);
 		__sync_fetch_and_add(&nr_direct_dispatches, 1);
 		dispatched = true;
 	}
@@ -1322,25 +1363,6 @@ static bool can_enqueue_to_cpu(const struct task_struct *p, s32 cpu)
 	return !is_cpu_busy(cpu);
 }
 
-/*
- * If tickless mode is enabled, check whether the task running on @cpu
- * needs to be preempted and, in that case, assign a regular time slice to
- * it.
- */
-static void preempt_curr(s32 cpu)
-{
-	struct task_struct *curr;
-
-	if (!tickless_sched)
-		return;
-
-	bpf_rcu_read_lock();
-	curr = scx_bpf_cpu_rq(cpu)->curr;
-	if (curr->scx.slice == SCX_SLICE_INF)
-		curr->scx.slice = task_slice(cpu);
-	bpf_rcu_read_unlock();
-}
-
 /*
  * Enqueue a task when running in round-robin mode.
  */
@@ -1363,13 +1385,13 @@ static void rr_enqueue(struct task_struct *p, struct task_ctx *tctx,
 			cpu = pick_idle_cpu(p, tctx, prev_cpu, 0, &is_idle);
 			if (is_idle) {
 				scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu,
-						   task_slice(cpu), enq_flags);
+						   task_slice(p, cpu), enq_flags);
 				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 				return;
 			}
 		}
 	}
-	scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, task_slice(prev_cpu), enq_flags);
+	scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, task_slice(p, prev_cpu), enq_flags);
 }
 
 /*
@@ -1383,12 +1405,6 @@ void BPF_STRUCT_OPS(flash_enqueue, struct task_struct *p, u64 enq_flags)
 	s32 prev_cpu = scx_bpf_task_cpu(p);
 	int node = __COMPAT_scx_bpf_cpu_node(prev_cpu);
 
-	/*
-	 * Task is going to be enqueued, so check whether its previously
-	 * used CPU needs to be preempted.
-	 */
-	preempt_curr(prev_cpu);
-
 	/*
 	 * Dispatch regular tasks to the shared DSQ.
 	 */
@@ -1425,14 +1441,14 @@ void BPF_STRUCT_OPS(flash_enqueue, struct task_struct *p, u64 enq_flags)
 	 */
 	if (!scx_bpf_task_running(p) && can_enqueue_to_cpu(p, prev_cpu)) {
 		scx_bpf_dsq_insert_vtime(p, cpu_to_dsq(prev_cpu),
-					 task_slice(prev_cpu), p->scx.dsq_vtime, enq_flags);
+					 task_slice(p, prev_cpu), p->scx.dsq_vtime, enq_flags);
 		__sync_fetch_and_add(&nr_shared_dispatches, 1);
 		scx_bpf_kick_cpu(prev_cpu, SCX_KICK_IDLE);
 
 		return;
 	}
 	scx_bpf_dsq_insert_vtime(p, node_to_dsq(node),
-				 task_slice(prev_cpu), p->scx.dsq_vtime, enq_flags);
+				 task_slice(p, prev_cpu), p->scx.dsq_vtime, enq_flags);
 	__sync_fetch_and_add(&nr_shared_dispatches, 1);
 
 	/*
@@ -1549,7 +1565,7 @@ void BPF_STRUCT_OPS(flash_dispatch, s32 cpu, struct task_struct *prev)
 	 * round on the same CPU.
 	 */
 	if (prev && keep_running(prev, cpu))
-		prev->scx.slice = task_slice(cpu);
+		prev->scx.slice = task_slice(prev, cpu);
 }
 
 /*
@@ -1956,6 +1972,53 @@ static void init_cpuperf_target(void)
 	scx_bpf_put_cpumask(online_cpumask);
 }
 
+/*
+ * Tickless timer used to preempt CPUs.
+ */
+static int tickless_timerfn(void *map, int *key, struct bpf_timer *timer)
+{
+	s32 cpu;
+	int err;
+
+	/*
+	 * Check if we need to preempt the running tasks.
+	 */
+	bpf_rcu_read_lock();
+	bpf_for(cpu, 0, nr_cpu_ids) {
+		struct task_struct *p;
+		struct rq *rq = scx_bpf_cpu_rq(cpu);
+
+		if (!rq)
+			continue;
+		/*
+		 * Ignore CPU if idle task is running.
+		 */
+		p = rq->curr;
+		if (p->flags & PF_IDLE)
+			continue;
+
+		/*
+		 * Ignore CPUs without any task waiting.
+		 */
+		if (!nr_tasks_waiting(cpu))
+			continue;
+
+		/*
+		 * Set a finite time slice to the running task, so that it
+		 * can be preempted.
+		 */
+		if (p->scx.slice == SCX_SLICE_INF)
+			p->scx.slice = slice_min;
+	}
+	bpf_rcu_read_unlock();
+
+	err = bpf_timer_start(timer, tick_interval_ns(), 0);
+	if (err)
+		scx_bpf_error("Failed to re-arm tickless timer");
+
+	return 0;
+}
+
 /*
  * Throttle timer used to inject idle time across all the CPUs.
  */
@@ -2124,6 +2187,25 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(flash_init)
 	if (err)
 		return err;
 
+	timer = bpf_map_lookup_elem(&tickless_timer, &key);
+	if (!timer) {
+		scx_bpf_error("Failed to lookup tickless timer");
+		return -ESRCH;
+	}
+
+	/*
+	 * Fire the tickless timer if tickless mode is enabled.
+	 */
+	if (tickless_sched) {
+		bpf_timer_init(timer, &tickless_timer, CLOCK_MONOTONIC);
+		bpf_timer_set_callback(timer, tickless_timerfn);
+		err = bpf_timer_start(timer, tick_interval_ns(), 0);
+		if (err) {
+			scx_bpf_error("Failed to arm tickless timer");
+			return err;
+		}
+	}
+
 	timer = bpf_map_lookup_elem(&throttle_timer, &key);
 	if (!timer) {
 		scx_bpf_error("Failed to lookup throttle timer");
@@ -2134,7 +2216,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(flash_init)
 	 * Fire the throttle timer if CPU throttling is enabled.
 	 */
 	if (throttle_ns) {
-		bpf_timer_init(timer, &throttle_timer, CLOCK_BOOTTIME);
+		bpf_timer_init(timer, &throttle_timer, CLOCK_MONOTONIC);
 		bpf_timer_set_callback(timer, throttle_timerfn);
 		err = bpf_timer_start(timer, slice_max, 0);
 		if (err) {
@@ -2153,7 +2235,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(flash_init)
 		return -ESRCH;
 	}
 
-	bpf_timer_init(timer, &numa_timer, CLOCK_BOOTTIME);
+	bpf_timer_init(timer, &numa_timer, CLOCK_MONOTONIC);
 	bpf_timer_set_callback(timer, numa_timerfn);
 	err = bpf_timer_start(timer, NSEC_PER_SEC, 0);
 	if (err) {
