@@ -208,17 +208,14 @@ static u64		cur_svc_time;
 const volatile u64	slice_min_ns = LAVD_SLICE_MIN_NS_DFL;
 const volatile u64	slice_max_ns = LAVD_SLICE_MAX_NS_DFL;
 
+static volatile u64	nr_cpus_big;
+
 /*
  * Include sub-modules
  */
 #include "util.bpf.c"
-#include "power.bpf.c"
-#include "introspec.bpf.c"
-#include "preempt.bpf.c"
-#include "lock.bpf.c"
 #include "idle.bpf.c"
 #include "balance.bpf.c"
-#include "sys_stat.bpf.c"
 #include "lat_cri.bpf.c"
 
 static void advance_cur_logical_clk(struct task_struct *p)
@@ -553,7 +550,10 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 			goto out;
 		}
 
-		dsq_id = cpdom_to_dsq(cpuc->cpdom_id);
+		if (per_cpu_dsq)
+			dsq_id = cpu_to_dsq(cpu_id);
+		else
+			dsq_id = cpdom_to_dsq(cpuc->cpdom_id);
 
 		if (!scx_bpf_dsq_nr_queued(dsq_id)) {
 			p->scx.dsq_vtime = calc_when_to_run(p, taskc);
@@ -1293,7 +1293,7 @@ void BPF_STRUCT_OPS(lavd_set_cpumask, struct task_struct *p,
 		return;
 	}
 
-	if (bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids)
+	if (bpf_cpumask_weight(p->cpus_ptr) != __nr_cpu_ids)
 		set_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
 	else
 		reset_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
@@ -1389,7 +1389,7 @@ static void init_task_ctx(struct task_struct *p, struct task_ctx *taskc)
 	u64 now = scx_bpf_now();
 
 	__builtin_memset(taskc, 0, sizeof(*taskc));
-	if (bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids)
+	if (bpf_cpumask_weight(p->cpus_ptr) != __nr_cpu_ids)
 		set_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
 	else
 		reset_task_flag(taskc, LAVD_FLAG_IS_AFFINITIZED);
@@ -1457,10 +1457,10 @@ static s32 init_cpdoms(u64 now)
 		/*
 		 * Create an associated DSQ on its associated NUMA domain.
 		 */
-		err = scx_bpf_create_dsq(cpdom_to_dsq(cpdomc->id), cpdomc->node_id);
+		err = scx_bpf_create_dsq(cpdom_to_dsq(cpdomc->id), cpdomc->numa_id);
 		if (err) {
 			scx_bpf_error("Failed to create a DSQ for cpdom %llu on NUMA node %d",
-				      cpdomc->id, cpdomc->node_id);
+				      cpdomc->id, cpdomc->numa_id);
 			return err;
 		}
 
@@ -1561,7 +1561,7 @@ static s32 init_per_cpu_ctx(u64 now)
 	 * Initilize CPU info
 	 */
 	one_little_capacity = LAVD_SCALE;
-	bpf_for(cpu, 0, nr_cpu_ids) {
+	bpf_for(cpu, 0, __nr_cpu_ids) {
 		if (cpu >= LAVD_CPU_ID_MAX)
 			break;
 
@@ -1667,6 +1667,7 @@ static s32 init_per_cpu_ctx(u64 now)
 						err = -ESRCH;
 						goto unlock_out;
 					}
+					cpuc->llc_id = cpdomc->llc_id;
 					cpuc->cpdom_id = cpdomc->id;
 					cpuc->cpdom_alt_id = cpdomc->alt_id;
 
@@ -1684,7 +1685,7 @@ static s32 init_per_cpu_ctx(u64 now)
 	/*
 	 * Print some useful informatin for debugging.
 	 */
-	bpf_for(cpu, 0, nr_cpu_ids) {
+	bpf_for(cpu, 0, __nr_cpu_ids) {
 		cpuc = get_cpu_ctx_id(cpu);
 		if (!cpuc) {
 			scx_bpf_error("Failed to lookup cpu_ctx: %d", cpu);
@@ -1710,7 +1711,7 @@ static int init_per_cpu_dsqs(void)
 	struct cpu_ctx *cpuc;
 	int cpu, err = 0;
 
-	bpf_for(cpu, 0, nr_cpu_ids) {
+	bpf_for(cpu, 0, __nr_cpu_ids) {
 		/*
 		 * Create Per-CPU DSQs on its associated NUMA domain.
 		 */
@@ -1726,10 +1727,13 @@ static int init_per_cpu_dsqs(void)
 			return -ESRCH;
 		}
 
-		err = scx_bpf_create_dsq(cpu_to_dsq(cpu), cpdomc->node_id);
+		if (is_smt_active && (cpu != get_primary_cpu(cpu)))
+			continue;
+
+		err = scx_bpf_create_dsq(cpu_to_dsq(cpu), cpdomc->numa_id);
 		if (err) {
 			scx_bpf_error("Failed to create a DSQ for cpu %d on NUMA node %d",
-				      cpu, cpdomc->node_id);
+				      cpu, cpdomc->numa_id);
 			return err;
 		}
 	}
