@@ -4,18 +4,22 @@
  * Author: Changwoo Min <changwoo@igalia.com>
  */
 
-/*
- * To be included to the main.bpf.c
- */
+#include <scx/common.bpf.h>
+#include "intf.h"
+#include "lavd.bpf.h"
+#include <errno.h>
+#include <stdbool.h>
+#include <bpf/bpf_core_read.h>
+#include <bpf/bpf_helpers.h>
+#include <bpf/bpf_tracing.h>
 
 /*
  * System-wide properties of CPUs
  */
-static bool		have_turbo_core;
-static bool		have_little_core;
+bool			have_turbo_core;
+bool			have_little_core;
 const volatile bool	is_smt_active;
 
-
 /*
  * CPU properties
  */
@@ -33,7 +37,7 @@ const volatile u8	cpu_turbo[LAVD_CPU_ID_MAX];
  * Compute domain properties
  */
 /* number of compute domains */
-static int		nr_cpdoms;
+int			nr_cpdoms;
 
 /* contexts for compute domains */
 struct cpdom_ctx	cpdom_ctxs[LAVD_CPDOM_MAX_NR];
@@ -67,16 +71,16 @@ volatile static int	pco_idx;
  * Big & LITTLE core's capacities
  */
 /* Total compute capacity of online CPUs. */
-static u64		total_capacity;
+u64		total_capacity;
 
 /* Capacity of one LITTLEst CPU. */
-static u64		one_little_capacity;
+u64		one_little_capacity;
 
 /* Big core's compute ratio among currently active cores scaled by 1024. */
-static u32		cur_big_core_scale;
+u32		cur_big_core_scale;
 
 /* Big core's compute ratio when all cores are active scaled by 1024. */
-static u32		default_big_core_scale;
+u32		default_big_core_scale;
 
 
 /*
@@ -86,17 +90,19 @@ static u64		LAVD_AP_LOW_CAP;
 static u64		LAVD_AP_HIGH_CAP;
 volatile int		power_mode;
 volatile u64		last_power_mode_clk;
-volatile u64		performance_mode_ns;
-volatile u64		balanced_mode_ns;
-volatile u64		powersave_mode_ns;
+volatile bool		is_powersave_mode;
 
-static void reset_suspended_duration(struct cpu_ctx *cpuc)
+__hidden
+int reset_suspended_duration(struct cpu_ctx *cpuc)
 {
 	if (cpuc->online_clk > cpuc->offline_clk)
 		cpuc->offline_clk = cpuc->online_clk;
+
+	return 0;
 }
 
-static u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
+__hidden
+u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
 {
 	/*
 	 * When a system is suspended, a task is also suspended in a running
@@ -116,8 +122,11 @@ static u64 get_suspended_duration_and_reset(struct cpu_ctx *cpuc)
 	return duration;
 }
 
-static bool is_perf_cri(struct task_ctx *taskc)
+bool is_perf_cri(struct task_ctx *taskc)
 {
+	if (unlikely(!taskc))
+		return false;
+
 	if (!have_little_core)
 		return true;
 
@@ -127,7 +136,8 @@ static bool is_perf_cri(struct task_ctx *taskc)
 	return test_task_flag(taskc, LAVD_FLAG_ON_BIG);
 }
 
-static const volatile u16 *get_cpu_order(void)
+__hidden
+const volatile u16 *get_cpu_order(void)
 {
 	int i = READ_ONCE(pco_idx);
 
@@ -197,7 +207,7 @@ static int calc_nr_active_cpus(void)
 		else
 			WRITE_ONCE(pco_idx, nr_pco_states - 1);
 
-		bpf_for(i, 0, nr_cpu_ids) {
+		bpf_for(i, 0, __nr_cpu_ids) {
 			if (i >= LAVD_CPU_ID_MAX)
 				break;
 
@@ -229,10 +239,11 @@ static int calc_nr_active_cpus(void)
 	}
 
 	/* Should not be here. */
-	return nr_cpu_ids;
+	return __nr_cpu_ids;
 }
 
-static void do_core_compaction(void)
+__weak
+int do_core_compaction(void)
 {
 	const volatile u16 *cpu_order;
 	struct cpu_ctx *cpuc;
@@ -267,7 +278,7 @@ static void do_core_compaction(void)
 	/*
 	 * Assign active and overflow cores.
 	 */
-	bpf_for(i, 0, nr_cpu_ids) {
+	bpf_for(i, 0, __nr_cpu_ids) {
 		if (i >= LAVD_CPU_ID_MAX)
 			break;
 
@@ -372,9 +383,11 @@ static void do_core_compaction(void)
 
 unlock_out:
 	bpf_rcu_read_unlock();
+
+	return 0;
 }
 
-static void update_power_mode_time(void)
+int update_power_mode_time(void)
 {
 	u64 now = scx_bpf_now();
 	u64 delta;
@@ -396,6 +409,8 @@ static void update_power_mode_time(void)
 		__sync_fetch_and_add(&powersave_mode_ns, delta);
 		break;
 	}
+	
+	return 0;
 }
 
 static int do_set_power_profile(s32 pm)
@@ -457,7 +472,8 @@ static int do_set_power_profile(s32 pm)
 	return 0;
 }
 
-static int do_autopilot(void)
+__weak
+int do_autopilot(void)
 {
 	/*
 	 * Calculate the required compute capacity from the scaled utilization.
@@ -490,7 +506,8 @@ static int do_autopilot(void)
 	return do_set_power_profile(LAVD_PM_PERFORMANCE);
 }
 
-static void update_thr_perf_cri(void)
+__weak
+int update_thr_perf_cri(void)
 {
 	u32 little_core_scale, delta, diff, thr;
 
@@ -502,7 +519,7 @@ static void update_thr_perf_cri(void)
 	 */
 	if (cur_big_core_scale == LAVD_SCALE) {
 		sys_stat.thr_perf_cri = 0;
-		return;
+		return 0;
 	}
 
 	/*
@@ -564,9 +581,12 @@ static void update_thr_perf_cri(void)
 	}
 
 	sys_stat.thr_perf_cri = thr;
+
+	return 0;
 }
 
-static int reinit_active_cpumask_for_performance(void)
+__weak
+int reinit_active_cpumask_for_performance(void)
 {
 	struct cpu_ctx *cpuc;
 	struct bpf_cpumask *active, *ovrflw;
@@ -599,7 +619,7 @@ static int reinit_active_cpumask_for_performance(void)
 	 * In a symmetric system, all online CPUs belong to the active set.
 	 */
 	if (have_little_core) {
-		bpf_for(cpu, 0, nr_cpu_ids) {
+		bpf_for(cpu, 0, __nr_cpu_ids) {
 			cpuc = get_cpu_ctx_id(cpu);
 			if (!cpuc)
 				continue;
@@ -632,7 +652,7 @@ static int reinit_active_cpumask_for_performance(void)
 
 		bpf_cpumask_clear(ovrflw);
 
-		bpf_for(cpu, 0, nr_cpu_ids) {
+		bpf_for(cpu, 0, __nr_cpu_ids) {
 			cpuc = get_cpu_ctx_id(cpu);
 			if (!cpuc || !cpuc->is_online)
 				continue;
@@ -673,7 +693,8 @@ unlock_out:
 	return err;
 }
 
-static void update_cpuperf_target(struct cpu_ctx *cpuc)
+__hidden
+int update_cpuperf_target(struct cpu_ctx *cpuc)
 {
 	u32 util, max_util, cpuperf_target;
 
@@ -698,16 +719,21 @@ static void update_cpuperf_target(struct cpu_ctx *cpuc)
 		scx_bpf_cpuperf_set(cpuc->cpu_id, cpuperf_target);
 		cpuc->cpuperf_cur = cpuperf_target;
 	}
+
+	return 0;
 }
 
-static void reset_cpuperf_target(struct cpu_ctx *cpuc)
+__hidden
+int reset_cpuperf_target(struct cpu_ctx *cpuc)
 {
 	if (!no_freq_scaling) {
 		cpuc->cpuperf_cur = 0;
 	}
+
+	return 0;
 }
 
-static u16 get_cpuperf_cap(s32 cpu)
+u16 get_cpuperf_cap(s32 cpu)
 {
 	const volatile u16 *cap;
 
@@ -719,7 +745,7 @@ static u16 get_cpuperf_cap(s32 cpu)
 	return 0;
 }
 
-static u64 scale_cap_freq(u64 dur, s32 cpu)
+u64 scale_cap_freq(u64 dur, s32 cpu)
 {
 	u64 cap, freq, scaled_dur;
 
@@ -746,13 +772,15 @@ static void do_update_autopilot_high_cap(void)
 	LAVD_AP_HIGH_CAP = c >> LAVD_SHIFT;
 }
 
-static void update_autopilot_high_cap(void)
+int update_autopilot_high_cap(void)
 {
 	if (no_use_em)
 		do_update_autopilot_high_cap();
+
+	return 0;
 }
 
-static void init_autopilot_caps(void)
+int init_autopilot_caps(void)
 {
 	if (no_use_em) {
 		/*
@@ -774,6 +802,8 @@ static void init_autopilot_caps(void)
 		LAVD_AP_LOW_CAP = pco_bounds[0];
 		LAVD_AP_HIGH_CAP = pco_bounds[i];
 	}
+
+	return 0;
 }
 
 SEC("syscall")
