@@ -17,16 +17,15 @@ private(LAVD) struct bpf_cpumask __kptr *little_cpumask; /* CPU mask for little
 private(LAVD) struct bpf_cpumask __kptr *active_cpumask; /* CPU mask for active CPUs */
 private(LAVD) struct bpf_cpumask __kptr *ovrflw_cpumask; /* CPU mask for overflow CPUs */
 
-const volatile u64	nr_cpu_ids;	/* maximum CPU IDs */
-static volatile u64	nr_cpus_onln;	/* current number of online CPUs */
-static volatile u64	nr_cpus_big;
+const volatile u64	nr_llcs;	/* number of LLC domains */
+const volatile u64	__nr_cpu_ids;	/* maximum CPU IDs */
+volatile u64		nr_cpus_onln;	/* current number of online CPUs */
 
-struct sys_stat	sys_stat;
+const volatile u32	cpu_sibling[LAVD_CPU_ID_MAX]; /* siblings for CPUs when SMT is active */
 
 /*
  * Options
  */
-volatile bool		is_powersave_mode;
 volatile bool		reinit_cpumask_for_performance;
 volatile bool		no_preemption;
 volatile bool		no_core_compaction;
@@ -37,7 +36,6 @@ const volatile bool	no_slice_boost;
 const volatile bool	per_cpu_dsq;
 const volatile bool	is_autopilot_on;
 const volatile u8	verbose;
-const volatile u8	preempt_shift;
 
 /*
  * Exit information
@@ -65,50 +63,28 @@ struct {
 } task_ctx_stor SEC(".maps");
 
 
-#define debugln(fmt, ...)						\
-({									\
-	if (verbose > 0)						\
-		bpf_printk("[%s:%d] " fmt, __func__, __LINE__,		\
-					##__VA_ARGS__);			\
-})
-
-#define traceln(fmt, ...)						\
-({									\
-	if (verbose > 1)						\
-		bpf_printk("[%s:%d] " fmt, __func__, __LINE__,		\
-					##__VA_ARGS__);			\
-})
-
-#ifndef min
-#define min(X, Y) (((X) < (Y)) ? (X) : (Y))
-#endif
-
-#ifndef max
-#define max(X, Y) (((X) < (Y)) ? (Y) : (X))
-#endif
-
-#ifndef clamp
-#define clamp(val, lo, hi) min(max(val, lo), hi)
-#endif
-
-static struct task_ctx *get_task_ctx(struct task_struct *p)
+__hidden
+struct task_ctx *get_task_ctx(struct task_struct *p)
 {
 	return bpf_task_storage_get(&task_ctx_stor, p, 0, 0);
 }
 
-static struct cpu_ctx *get_cpu_ctx(void)
+__hidden
+struct cpu_ctx *get_cpu_ctx(void)
 {
 	const u32 idx = 0;
 	return bpf_map_lookup_elem(&cpu_ctx_stor, &idx);
 }
 
-static struct cpu_ctx *get_cpu_ctx_id(s32 cpu_id)
+__hidden
+struct cpu_ctx *get_cpu_ctx_id(s32 cpu_id)
 {
 	const u32 idx = 0;
 	return bpf_map_lookup_percpu_elem(&cpu_ctx_stor, &idx, cpu_id);
 }
 
-static struct cpu_ctx *get_cpu_ctx_task(const struct task_struct *p)
+__hidden
+struct cpu_ctx *get_cpu_ctx_task(const struct task_struct *p)
 {
 	return get_cpu_ctx_id(scx_bpf_task_cpu(p));
 }
@@ -170,52 +146,61 @@ static bool is_pinned(const struct task_struct *p)
 	return p->nr_cpus_allowed == 1;
 }
 
-static inline bool test_task_flag(struct task_ctx *taskc, u64 flag)
+__hidden
+bool test_task_flag(struct task_ctx *taskc, u64 flag)
 {
 	return (taskc->flags & flag) == flag;
 }
 
-static inline void set_task_flag(struct task_ctx *taskc, u64 flag)
+__hidden
+void set_task_flag(struct task_ctx *taskc, u64 flag)
 {
 	taskc->flags |= flag;
 }
 
-static inline void reset_task_flag(struct task_ctx *taskc, u64 flag)
+__hidden
+void reset_task_flag(struct task_ctx *taskc, u64 flag)
 {
 	taskc->flags &= ~flag;
 }
 
-static inline bool test_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
+__hidden
+inline bool test_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
 {
 	return (cpuc->flags & flag) == flag;
 }
 
-static inline void set_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
+__hidden
+inline void set_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
 {
 	cpuc->flags |= flag;
 }
 
-static inline void reset_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
+__hidden
+inline void reset_cpu_flag(struct cpu_ctx *cpuc, u64 flag)
 {
 	cpuc->flags &= ~flag;
 }
 
-static bool is_lat_cri(struct task_ctx *taskc)
+__hidden
+bool is_lat_cri(struct task_ctx *taskc)
 {
 	return taskc->lat_cri >= sys_stat.avg_lat_cri;
 }
 
-static bool is_lock_holder(struct task_ctx *taskc)
+__hidden
+bool is_lock_holder(struct task_ctx *taskc)
 {
 	return test_task_flag(taskc, LAVD_FLAG_FUTEX_BOOST);
 }
 
-static bool is_lock_holder_running(struct cpu_ctx *cpuc)
+__hidden
+bool is_lock_holder_running(struct cpu_ctx *cpuc)
 {
 	return test_cpu_flag(cpuc, LAVD_FLAG_FUTEX_BOOST);
 }
 
-static bool have_scheduled(struct task_ctx *taskc)
+bool have_scheduled(struct task_ctx *taskc)
 {
 	/*
 	 * If task's time slice hasn't been updated, that means the task has
@@ -224,12 +209,12 @@ static bool have_scheduled(struct task_ctx *taskc)
 	return taskc->slice != 0;
 }
 
-static bool can_boost_slice(void)
+bool can_boost_slice(void)
 {
 	return sys_stat.nr_queued_task <= sys_stat.nr_active;
 }
 
-static u16 get_nice_prio(struct task_struct *p)
+u16 get_nice_prio(struct task_struct __arg_trusted *p)
 {
 	u16 prio = p->static_prio - MAX_RT_PRIO; /* [0, 40) */
 	return prio;
@@ -260,7 +245,7 @@ static void set_on_core_type(struct task_ctx *taskc,
 	struct cpu_ctx *cpuc;
 	int cpu;
 
-	bpf_for(cpu, 0, nr_cpu_ids) {
+	bpf_for(cpu, 0, __nr_cpu_ids) {
 		if (!bpf_cpumask_test_cpu(cpu, cpumask))
 			continue;
 
@@ -304,4 +289,25 @@ bool __attribute__ ((noinline)) prob_x_out_of_y(u32 x, u32 y)
 	r = bpf_get_prandom_u32() % y;
 	return r < x;
 }
+/*
+ * We define the primary cpu in the physical core as the lowest logical cpu id.
+ */
+u32 __attribute__ ((noinline)) get_primary_cpu(u32 cpu) {
+	const volatile u32 *sibling;
 
+	if (!is_smt_active)
+		return cpu;
+
+	sibling = MEMBER_VPTR(cpu_sibling, [cpu]);
+	if (!sibling) {
+		debugln("Infeasible CPU id: %d", cpu);
+		return cpu;
+	}
+
+	return ((cpu < *sibling) ? cpu : *sibling);
+}
+
+u32 cpu_to_dsq(u32 cpu)
+{
+	return (get_primary_cpu(cpu)) | LAVD_DSQ_TYPE_CPU << LAVD_DSQ_TYPE_SHFT;
+}
