@@ -47,10 +47,11 @@ use scx_stats::prelude::*;
 use scx_utils::autopower::{fetch_power_profile, PowerProfile};
 use scx_utils::build_id;
 use scx_utils::compat;
+use scx_utils::libbpf_clap_opts::LibbpfOpts;
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
-use scx_utils::set_rlimit_infinity;
+use scx_utils::try_set_rlimit_infinity;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
 use scx_utils::EnergyModel;
@@ -62,6 +63,7 @@ use stats::StatsReq;
 use stats::StatsRes;
 use stats::SysStats;
 
+const SCHEDULER_NAME: &str = "scx_lavd";
 /// scx_lavd: Latency-criticality Aware Virtual Deadline (LAVD) scheduler
 ///
 /// The rust part is minimal. It processes command line options and logs out
@@ -200,6 +202,9 @@ struct Opts {
     /// Show descriptions for statistics.
     #[clap(long)]
     help_stats: bool,
+
+    #[clap(flatten, next_help_heading = "Libbpf Options")]
+    pub libbpf: LibbpfOpts,
 }
 
 impl Opts {
@@ -338,16 +343,15 @@ impl<'a> Scheduler<'a> {
             );
         }
 
-        // Increase MEMLOCK size since the BPF scheduler might use
-        // more than the current limit
-        set_rlimit_infinity();
+        try_set_rlimit_infinity();
 
         // Open the BPF prog first for verification.
         let mut skel_builder = BpfSkelBuilder::default();
         skel_builder.obj_builder.debug(opts.verbose > 0);
         init_libbpf_logging(Some(PrintLevel::Debug));
 
-        let mut skel = scx_ops_open!(skel_builder, open_object, lavd_ops)?;
+        let open_opts = opts.libbpf.clone().into_bpf_open_opts();
+        let mut skel = scx_ops_open!(skel_builder, open_object, lavd_ops, open_opts)?;
 
         // Enable futex tracing using ftrace if available. If the ftrace is not
         // available, use tracepoint, which is known to be slower than ftrace.
@@ -437,11 +441,13 @@ impl<'a> Scheduler<'a> {
     fn init_cpus(skel: &mut OpenBpfSkel, order: &CpuOrder) {
         debug!("{:#?}", order);
 
-        // Initialize CPU capacity.
+        // Initialize CPU capacity and sibling
         for cpu in order.cpuids.iter() {
             skel.maps.rodata_data.as_mut().unwrap().cpu_capacity[cpu.cpu_adx] = cpu.cpu_cap as u16;
             skel.maps.rodata_data.as_mut().unwrap().cpu_big[cpu.cpu_adx] = cpu.big_core as u8;
             skel.maps.rodata_data.as_mut().unwrap().cpu_turbo[cpu.cpu_adx] = cpu.turbo_core as u8;
+            skel.maps.rodata_data.as_mut().unwrap().cpu_sibling[cpu.cpu_adx] =
+                cpu.cpu_sibling as u32;
         }
 
         // Initialize performance vs. CPU order table.
@@ -486,7 +492,8 @@ impl<'a> Scheduler<'a> {
             skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].id = v.cpdom_id as u64;
             skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].alt_id =
                 v.cpdom_alt_id.get() as u64;
-            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].node_id = k.node_adx as u8;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].numa_id = k.numa_adx as u8;
+            skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].llc_id = k.llc_adx as u8;
             skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].is_big = k.is_big as u8;
             skel.maps.bss_data.as_mut().unwrap().cpdom_ctxs[v.cpdom_id].is_valid = 1;
             for cpu_id in v.cpu_ids.iter() {
@@ -522,7 +529,8 @@ impl<'a> Scheduler<'a> {
         bss_data.no_freq_scaling = opts.no_freq_scaling;
         bss_data.is_powersave_mode = opts.powersave;
         let rodata = skel.maps.rodata_data.as_mut().unwrap();
-        rodata.nr_cpu_ids = *NR_CPU_IDS as u64;
+        rodata.nr_llcs = order.nr_llcs as u64;
+        rodata.__nr_cpu_ids = *NR_CPU_IDS as u64;
         rodata.is_smt_active = order.smt_enabled;
         rodata.is_autopilot_on = opts.autopilot;
         rodata.verbose = opts.verbose;
@@ -811,6 +819,8 @@ impl<'a> Scheduler<'a> {
 
 impl Drop for Scheduler<'_> {
     fn drop(&mut self) {
+        info!("Unregister {SCHEDULER_NAME} scheduler");
+
         if let Some(struct_ops) = self.struct_ops.take() {
             drop(struct_ops);
         }
