@@ -76,6 +76,7 @@ pub enum LayerMatch {
     CgroupPrefix(String),
     CgroupSuffix(String),
     CgroupContains(String),
+    CgroupRegex(String),
     CommPrefix(String),
     CommPrefixExclude(String),
     PcommPrefix(String),
@@ -96,6 +97,7 @@ pub enum LayerMatch {
     UsedGpuTid(bool),
     UsedGpuPid(bool),
     AvgRuntime(u64, u64),
+    HintEquals(u64),
 }
 
 #[derive(Clone, Debug, Serialize, Deserialize)]
