@@ -29,7 +29,10 @@ use log::info;
 use log::trace;
 use log::warn;
 use scx_stats::prelude::*;
+use scx_utils::build_id;
+use scx_utils::compat;
 use scx_utils::init_libbpf_logging;
+use scx_utils::libbpf_clap_opts::LibbpfOpts;
 use scx_utils::scx_enums;
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
@@ -44,6 +47,7 @@ use scx_utils::NR_CPUS_POSSIBLE;
 use stats::CellMetrics;
 use stats::Metrics;
 
+const SCHEDULER_NAME: &str = "scx_mitosis";
 const MAX_CELLS: usize = bpf_intf::consts_MAX_CELLS as usize;
 const NR_CSTATS: usize = bpf_intf::cell_stat_idx_NR_CSTATS as usize;
 
@@ -81,6 +85,13 @@ struct Opts {
     /// is not launched.
     #[clap(long)]
     monitor: Option<f64>,
+
+    /// Print scheduler version and exit.
+    #[clap(short = 'V', long, action = clap::ArgAction::SetTrue)]
+    version: bool,
+
+    #[clap(flatten, next_help_heading = "Libbpf Options")]
+    pub libbpf: LibbpfOpts,
 }
 
 // The subset of cstats we care about.
@@ -108,6 +119,7 @@ struct Scheduler<'a> {
     prev_cell_stats: [[u64; NR_CSTATS]; MAX_CELLS],
     metrics: Metrics,
     stats_server: StatsServer<(), Metrics>,
+    last_configuration_seq: Option<u32>,
 }
 
 struct DistributionStats {
@@ -153,7 +165,13 @@ impl<'a> Scheduler<'a> {
         let mut skel_builder = BpfSkelBuilder::default();
         skel_builder.obj_builder.debug(opts.verbose > 1);
         init_libbpf_logging(None);
-        let mut skel = scx_ops_open!(skel_builder, open_object, mitosis)?;
+        info!(
+            "Running scx_mitosis (build ID: {})",
+            build_id::full_version(env!("CARGO_PKG_VERSION"))
+        );
+
+        let open_opts = opts.libbpf.clone().into_bpf_open_opts();
+        let mut skel = scx_ops_open!(skel_builder, open_object, mitosis, open_opts)?;
 
         skel.struct_ops.mitosis_mut().exit_dump_len = opts.exit_dump_len;
 
@@ -164,6 +182,11 @@ impl<'a> Scheduler<'a> {
             skel.maps.rodata_data.as_mut().unwrap().all_cpus[cpu / 8] |= 1 << (cpu % 8);
         }
 
+        match *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP {
+            0 => info!("Kernel does not support queued wakeup optimization."),
+            v => skel.struct_ops.mitosis_mut().flags |= v,
+        }
+
         let skel = scx_ops_load!(skel, mitosis, uei)?;
 
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
@@ -175,13 +198,14 @@ impl<'a> Scheduler<'a> {
             prev_cell_stats: [[0; NR_CSTATS]; MAX_CELLS],
             metrics: Metrics::default(),
             stats_server,
+            last_configuration_seq: None,
         })
     }
 
     fn run(&mut self, shutdown: Arc<AtomicBool>) -> Result<UserExitInfo> {
         let struct_ops = scx_ops_attach!(self.skel, mitosis)?;
 
-        info!("Layered Scheduler Attached. Run `scx_mitosis --monitor` for metrics.");
+        info!("Mitosis Scheduler Attached. Run `scx_mitosis --monitor` for metrics.");
 
         let (res_ch, req_ch) = self.stats_server.channels();
 
@@ -196,6 +220,7 @@ impl<'a> Scheduler<'a> {
             }
         }
         drop(struct_ops);
+        info!("Unregister {SCHEDULER_NAME} scheduler");
         uei_report!(&self.skel, uei)
     }
 
@@ -409,18 +434,32 @@ impl<'a> Scheduler<'a> {
     }
 
     fn refresh_bpf_cells(&mut self) -> Result<()> {
+        let applied_configuration = unsafe {
+            std::ptr::read_volatile(
+                &self
+                    .skel
+                    .maps
+                    .bss_data
+                    .as_ref()
+                    .unwrap()
+                    .applied_configuration_seq as *const u32,
+            )
+        };
+        if self
+            .last_configuration_seq
+            .is_some_and(|seq| applied_configuration == seq)
+        {
+            return Ok(());
+        }
         // collect all cpus per cell.
         let mut cell_to_cpus: HashMap<u32, Cpumask> = HashMap::new();
         let cpu_ctxs = read_cpu_ctxs(&self.skel)?;
         for (i, cpu_ctx) in cpu_ctxs.iter().enumerate() {
             cell_to_cpus
                 .entry(cpu_ctx.cell)
-                .and_modify(|mask| mask.set_cpu(i).expect("set cpu in existing mask"))
-                .or_insert_with(|| {
-                    let mut mask = Cpumask::new();
-                    mask.set_cpu(i).expect("set cpu in new mask");
-                    mask
-                });
+                .or_insert_with(|| Cpumask::new())
+                .set_cpu(i)
+                .expect("set cpu in existing mask");
         }
 
         // Create cells we don't have yet, drop cells that are no longer in use.
@@ -430,13 +469,17 @@ impl<'a> Scheduler<'a> {
         for i in 0..MAX_CELLS {
             let cell_idx = i as u32;
             let bpf_cell = cells[i];
-            if bpf_cell.in_use > 0 {
-                self.cells.entry(cell_idx).or_insert(Cell {
-                    cpus: cell_to_cpus
-                        .get(&cell_idx)
-                        .expect("missing cell in cpu map")
-                        .clone(),
-                });
+            let in_use = unsafe { std::ptr::read_volatile(&bpf_cell.in_use as *const u32) };
+            if in_use > 0 {
+                self.cells
+                    .entry(cell_idx)
+                    .or_insert_with(|| Cell {
+                        cpus: Cpumask::new(),
+                    })
+                    .cpus = cell_to_cpus
+                    .get(&cell_idx)
+                    .expect("missing cell in cpu map")
+                    .clone();
                 self.metrics.cells.insert(cell_idx, CellMetrics::default());
             } else {
                 self.cells.remove(&cell_idx);
@@ -444,6 +487,8 @@ impl<'a> Scheduler<'a> {
             }
         }
 
+        self.last_configuration_seq = Some(applied_configuration);
+
         Ok(())
     }
 }
@@ -467,6 +512,14 @@ fn read_cpu_ctxs(skel: &BpfSkel) -> Result<Vec<bpf_intf::cpu_ctx>> {
 fn main() -> Result<()> {
     let opts = Opts::parse();
 
+    if opts.version {
+        println!(
+            "scx_mitosis {}",
+            build_id::full_version(env!("CARGO_PKG_VERSION"))
+        );
+        return Ok(());
+    }
+
     let llv = match opts.verbose {
         0 => simplelog::LevelFilter::Info,
         1 => simplelog::LevelFilter::Debug,
