@@ -80,30 +80,30 @@ const volatile struct {
 };
 
 const volatile struct {
-	u64 slack_factor;
 	u64 backoff_ns;
-	u64 min_nr_queued_pick2;
-	bool max_dsq_pick2;
-	bool eager_load_balance;
-
 	u64 dispatch_lb_busy;
-	bool dispatch_pick2_disable;
-	bool dispatch_lb_interactive;
-
+	u64 min_llc_runs_pick2;
+	u64 min_nr_queued_pick2;
+	u64 slack_factor;
 	u64 wakeup_lb_busy;
+
+	bool dispatch_lb_interactive;
+	bool dispatch_pick2_disable;
+	bool eager_load_balance;
+	bool max_dsq_pick2;
 	bool wakeup_llc_migrations;
 } lb_config = {
-	.slack_factor = LOAD_BALANCE_SLACK,
 	.backoff_ns = 5LLU * NSEC_PER_MSEC,
-	.min_nr_queued_pick2 = 10,
-	.max_dsq_pick2 = false,
-	.eager_load_balance = true,
-
 	.dispatch_lb_busy = 75,
-	.dispatch_pick2_disable = false,
-	.dispatch_lb_interactive = false,
-
+	.min_llc_runs_pick2 = 4,
+	.min_nr_queued_pick2 = 10,
+	.slack_factor = LOAD_BALANCE_SLACK,
 	.wakeup_lb_busy = 90,
+
+	.dispatch_lb_interactive = false,
+	.dispatch_pick2_disable = false,
+	.eager_load_balance = true,
+	.max_dsq_pick2 = false,
 	.wakeup_llc_migrations = false,
 };
 
@@ -113,15 +113,17 @@ const volatile struct {
 	u64 dsq_shift;
 	u32 interactive_ratio;
 	u32 saturated_percent;
+	u32 sched_mode;
 
 	bool atq_enabled;
 	bool cpu_priority;
+	bool task_slice;
 	bool freq_control;
 	bool interactive_sticky;
 	bool keep_running_enabled;
 	bool kthreads_local;
-	bool select_idle_in_enqueue;
 } p2dq_config = {
+	.sched_mode = MODE_DEFAULT,
 	.nr_dsqs_per_llc = 3,
 	.init_dsq_index = 0,
 	.dsq_shift = 2,
@@ -130,11 +132,11 @@ const volatile struct {
 
 	.atq_enabled = false,
 	.cpu_priority = false,
+	.task_slice = true,
 	.freq_control = false,
 	.interactive_sticky = false,
 	.keep_running_enabled = true,
 	.kthreads_local = true,
-	.select_idle_in_enqueue = true,
 };
 
 const volatile u32 debug = 2;
@@ -144,7 +146,7 @@ extern const volatile u32 nr_cpu_ids;
 const u64 lb_timer_intvl_ns = 250LLU * NSEC_PER_MSEC;
 
 static u32 llc_lb_offset = 1;
-static u32 min_llc_runs_pick2 = 1;
+static u64 min_llc_runs_pick2 = 1;
 static bool saturated = false;
 
 u64 llc_ids[MAX_LLCS];
@@ -154,7 +156,6 @@ u64 big_core_ids[MAX_CPUS];
 u64 dsq_time_slices[MAX_DSQS_PER_LLC];
 
 u64 min_slice_ns = 500;
-u32 sched_mode = MODE_PERFORMANCE;
 
 private(A) struct bpf_cpumask __kptr *big_cpumask;
 
@@ -261,6 +262,11 @@ static u64 task_dsq_slice_ns(struct task_struct *p, int dsq_index)
 	return task_slice_ns(p, dsq_time_slice(dsq_index));
 }
 
+static void task_refresh_llc_runs(task_ctx *taskc)
+{
+	taskc->llc_runs = min_llc_runs_pick2;
+}
+
 static u64 llc_nr_queued(struct llc_ctx *llcx)
 {
 	u64 nr_queued = scx_bpf_dsq_nr_queued(llcx->dsq);
@@ -449,16 +455,16 @@ static bool can_migrate(task_ctx *taskc, struct llc_ctx *llcx)
 	    (!lb_config.dispatch_lb_interactive && taskc->interactive))
 		return false;
 
-	if (saturated)
-		return true;
-
 	if (lb_config.max_dsq_pick2 &&
 	    taskc->dsq_index != p2dq_config.nr_dsqs_per_llc - 1)
 		return false;
 
-	if (taskc->llc_runs < min_llc_runs_pick2)
+	if (taskc->llc_runs > 0)
 		return false;
 
+	if (saturated)
+		return true;
+
 	return llcx->saturated;
 }
 
@@ -742,8 +748,44 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 		goto found_cpu;
 	}
 
+	if (p2dq_config.sched_mode == MODE_PERF &&
+	    topo_config.has_little_cores &&
+	    llcx->big_cpumask) {
+		cpu = __pick_idle_cpu(llcx->big_cpumask,
+				      SCX_PICK_IDLE_CORE);
+		if (cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+		if (llcx->big_cpumask) {
+			cpu = __pick_idle_cpu(llcx->big_cpumask, 0);
+			if (cpu >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		}
+	}
+
+	if (p2dq_config.sched_mode == MODE_EFFICIENCY &&
+	    topo_config.has_little_cores &&
+	    llcx->little_cpumask) {
+		cpu = __pick_idle_cpu(llcx->little_cpumask, SCX_PICK_IDLE_CORE);
+		if (cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+		if (llcx->little_cpumask) {
+			cpu = __pick_idle_cpu(llcx->little_cpumask, 0);
+			if (cpu >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		}
+	}
+
+
 	if (llcx->lb_llc_id < MAX_LLCS &&
-	    taskc->llc_runs > min_llc_runs_pick2) {
+	    taskc->llc_runs == 0) {
 		u32 target_llc_id = llcx->lb_llc_id;
 		llcx->lb_llc_id = MAX_LLCS;
 		if (!(llcx = lookup_llc_ctx(target_llc_id)))
@@ -798,6 +840,14 @@ static s32 pick_idle_cpu(struct task_struct *p, task_ctx *taskc,
 		goto found_cpu;
 	}
 
+	if (saturated && taskc->llc_runs == 0 && llcx->node_cpumask) {
+		cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->node_cpumask), 0);
+		if (cpu >= 0) {
+			*is_idle = true;
+			goto found_cpu;
+		}
+	}
+
 	// Couldn't find anything idle just return something in the local LLC
 	if (taskc->interactive && llcx->cpumask)
 		cpu = bpf_cpumask_any_distribute(cast_mask(llcx->cpumask));
@@ -940,8 +990,7 @@ static void async_p2dq_enqueue(struct enqueue_promise *ret,
 	}
 
 	// If an idle CPU hasn't been found in select_cpu find one now
-	if (p2dq_config.select_idle_in_enqueue &&
-	    !__COMPAT_is_enq_cpu_selected(enq_flags)) {
+	if (!__COMPAT_is_enq_cpu_selected(enq_flags)) {
 		cpu = pick_idle_cpu(p,
 				    taskc,
 				    cpu,
@@ -1144,13 +1193,16 @@ static int p2dq_running_impl(struct task_struct *p)
 		return -EINVAL;
 
 	if (taskc->llc_id != cpuc->llc_id) {
-		taskc->llc_runs = 0;
+		task_refresh_llc_runs(taskc);
 		stat_inc(P2DQ_STAT_LLC_MIGRATION);
 		trace("RUNNING %d cpu %d->%d llc %d->%d",
 		      p->pid, cpuc->id, task_cpu,
 		      taskc->llc_id, llcx->id);
 	} else {
-		taskc->llc_runs += 1;
+		if (taskc->llc_runs == 0)
+			task_refresh_llc_runs(taskc);
+		else
+			taskc->llc_runs -= 1;
 	}
 	if (taskc->node_id != cpuc->node_id) {
 		stat_inc(P2DQ_STAT_NODE_MIGRATION);
@@ -1273,7 +1325,15 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 		if (p->scx.weight < 100 && taskc->dsq_index > 1)
 			taskc->dsq_index = 1;
 
-		taskc->slice_ns = task_dsq_slice_ns(p, taskc->dsq_index);
+		if (p2dq_config.task_slice) {
+			if (used >= ((7 * last_dsq_slice_ns) / 8)) {
+				taskc->slice_ns = clamp_slice((5 * taskc->slice_ns) >> 2);
+			} else if (used < last_dsq_slice_ns / 2) {
+				taskc->slice_ns = clamp_slice((7 * taskc->slice_ns) >> 3);
+			}
+		} else {
+			taskc->slice_ns = task_dsq_slice_ns(p, taskc->dsq_index);
+		}
 		taskc->last_run_started = 0;
 		taskc->interactive = is_interactive(taskc);
 	}
@@ -1403,6 +1463,12 @@ static int dispatch_pick_two(s32 cpu, struct llc_ctx *cur_llcx, struct cpu_ctx *
 
 		if (consume_llc(second))
 			return 0;
+
+		// If the system is saturated then be aggressive in trying to load balance.
+		if (topo_config.nr_llcs > 2 &&
+		    (first = rand_llc_ctx()) &&
+		    consume_llc(first))
+			return 0;
 	}
 
 	return 0;
@@ -1599,11 +1665,11 @@ void BPF_STRUCT_OPS(p2dq_update_idle, s32 cpu, bool idle)
 	percent_idle = idle_cpu_percent(idle_cpumask);
 	saturated = percent_idle < p2dq_config.saturated_percent;
 
-	if (saturated)
-		min_llc_runs_pick2 = 0;
-	else {
+	if (saturated) {
+		min_llc_runs_pick2 = min(2, lb_config.min_llc_runs_pick2);
+	} else {
 		u32 llc_scaler = log2_u32(topo_config.nr_llcs);
-		min_llc_runs_pick2 = log2_u32(percent_idle) + llc_scaler;
+		min_llc_runs_pick2 = min(log2_u32(percent_idle) + llc_scaler, lb_config.min_llc_runs_pick2);
 	}
 
 	if (!(llcx = lookup_cpu_llc_ctx(cpu))) {
@@ -1708,6 +1774,7 @@ static s32 p2dq_init_task_impl(struct task_struct *p, struct scx_init_task_args
 			  p->nr_cpus_allowed == topo_config.nr_cpus;
 	taskc->interactive = is_interactive(taskc);
 	p->scx.dsq_vtime = llcx->vtime;
+	task_refresh_llc_runs(taskc);
 
 	// When a task is initialized set the DSQ id to invalid. This causes
 	// the task to be randomized on a LLC.
