@@ -4,6 +4,9 @@
 
 #include <string.h>
 
+#include <scx/common.bpf.h>
+#include <lib/sdt_task_defs.h>
+
 #include "main.bpf.c"
 
 // per thread globals because the Rust test driver has multiple threads
@@ -58,7 +61,6 @@ SCX_TEST(test_pick_idle_cpu)
 	struct task_struct p = { 0 };
 	task_ctx my_taskc = { 0 };
 	struct cpumask llc_cpumask = { 0 };
-	u64 index = 0;
 	s32 idle_cpu;
 	bool is_idle = false;
 
