@@ -24,6 +24,7 @@ use scx_stats::prelude::*;
 use scx_utils::build_id;
 use scx_utils::compat;
 use scx_utils::init_libbpf_logging;
+use scx_utils::libbpf_clap_opts::LibbpfOpts;
 use scx_utils::pm::{cpu_idle_resume_latency_supported, update_cpu_idle_resume_latency};
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
@@ -61,6 +62,7 @@ use scx_p2dq::bpf_skel::*;
 use scx_p2dq::SchedulerOpts;
 use scx_p2dq::TOPO;
 
+const SCHEDULER_NAME: &str = "scx_p2dq";
 /// scx_p2dq: A pick 2 dumb queuing load balancing scheduler.
 ///
 /// The BPF part does simple vtime or round robin scheduling in each domain
@@ -88,6 +90,9 @@ struct CliOpts {
 
     #[clap(flatten)]
     pub sched: SchedulerOpts,
+
+    #[clap(flatten, next_help_heading = "Libbpf Options")]
+    pub libbpf: LibbpfOpts,
 }
 
 struct Scheduler<'a> {
@@ -101,6 +106,7 @@ struct Scheduler<'a> {
 impl<'a> Scheduler<'a> {
     fn init(
         opts: &SchedulerOpts,
+        libbpf_ops: &LibbpfOpts,
         open_object: &'a mut MaybeUninit<OpenObject>,
         verbose: u8,
     ) -> Result<Self> {
@@ -112,7 +118,15 @@ impl<'a> Scheduler<'a> {
             "Running scx_p2dq (build ID: {})",
             build_id::full_version(env!("CARGO_PKG_VERSION"))
         );
-        let mut open_skel = scx_ops_open!(skel_builder, open_object, p2dq).unwrap();
+        let open_opts = libbpf_ops.clone().into_bpf_open_opts();
+        let mut open_skel = scx_ops_open!(skel_builder, open_object, p2dq, open_opts).context(
+            "Failed to open BPF object. This can be caused by a mismatch between the kernel \
+            version and the BPF object, permission or other libbpf issues. Try running `dmesg \
+            | grep bpf` to see if there are any error messages related to the BPF object. See \
+            the LibbpfOptions section in the help for more information on configuration related \
+            to this issue or file an issue on the scx repo if the problem persists. \
+            https://github.com/sched-ext/scx/issues/new?labels=scx_p2dq&title=scx_p2dq:%20New%20Issue&assignees=hodgesds&body=Kernel%20version:%20(fill%20me%20out)%0ADistribution:%20(fill%20me%20out)%0AHardware:%20(fill%20me%20out)%0A%0AIssue:%20(fill%20me%20out)"
+        )?;
         scx_p2dq::init_open_skel!(&mut open_skel, opts, verbose)?;
 
         if opts.queued_wakeup {
@@ -121,7 +135,6 @@ impl<'a> Scheduler<'a> {
         open_skel.struct_ops.p2dq_mut().flags |= *compat::SCX_OPS_KEEP_BUILTIN_IDLE;
 
         let mut skel = scx_ops_load!(open_skel, p2dq, uei)?;
-
         scx_p2dq::init_skel!(&mut skel);
 
         let stats_server = StatsServer::new(stats::server_data()).launch()?;
@@ -153,7 +166,6 @@ impl<'a> Scheduler<'a> {
             atq_reenq: stats[stat_idx_P2DQ_STAT_ATQ_REENQ as usize],
             direct: stats[stat_idx_P2DQ_STAT_DIRECT as usize],
             idle: stats[stat_idx_P2DQ_STAT_IDLE as usize],
-            sched_mode: self.skel.maps.bss_data.as_ref().unwrap().sched_mode,
             dsq_change: stats[stat_idx_P2DQ_STAT_DSQ_CHANGE as usize],
             same_dsq: stats[stat_idx_P2DQ_STAT_DSQ_SAME as usize],
             keep: stats[stat_idx_P2DQ_STAT_KEEP as usize],
@@ -340,6 +352,8 @@ impl<'a> Scheduler<'a> {
 
 impl Drop for Scheduler<'_> {
     fn drop(&mut self) {
+        info!("Unregister {SCHEDULER_NAME} scheduler");
+
         if let Some(struct_ops) = self.struct_ops.take() {
             drop(struct_ops);
         }
@@ -391,10 +405,7 @@ fn main() -> Result<()> {
                     debug!("stats monitor thread finished successfully")
                 }
                 Err(error_object) => {
-                    warn!(
-                        "stats monitor thread finished because of an error {}",
-                        error_object
-                    )
+                    warn!("stats monitor thread finished because of an error {error_object}")
                 }
             }
         });
@@ -407,19 +418,17 @@ fn main() -> Result<()> {
     if let Some(idle_resume_us) = opts.sched.idle_resume_us {
         if !cpu_idle_resume_latency_supported() {
             warn!("idle resume latency not supported");
-        } else {
-            if idle_resume_us > 0 {
-                info!("Setting idle QoS to {}us", idle_resume_us);
-                for cpu in TOPO.all_cpus.values() {
-                    update_cpu_idle_resume_latency(cpu.id, idle_resume_us.try_into().unwrap())?;
-                }
+        } else if idle_resume_us > 0 {
+            info!("Setting idle QoS to {idle_resume_us}us");
+            for cpu in TOPO.all_cpus.values() {
+                update_cpu_idle_resume_latency(cpu.id, idle_resume_us.try_into().unwrap())?;
             }
         }
     }
 
     let mut open_object = MaybeUninit::uninit();
     loop {
-        let mut sched = Scheduler::init(&opts.sched, &mut open_object, opts.verbose)?;
+        let mut sched = Scheduler::init(&opts.sched, &opts.libbpf, &mut open_object, opts.verbose)?;
         sched.start()?;
 
         if !sched.run(shutdown.clone())?.should_restart() {
