@@ -137,6 +137,7 @@ use std::collections::VecDeque;
 use std::fmt;
 use std::sync::Arc;
 
+use anyhow::anyhow;
 use anyhow::bail;
 use anyhow::Result;
 use log::debug;
@@ -533,14 +534,16 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
         let num_numa_nodes = self.dom_group.nr_nodes();
         let numa_load_avg = total_load / num_numa_nodes as f64;
 
-        let mut nodes: Vec<NumaNode> = Vec::with_capacity(num_numa_nodes);
-        for id in 0..num_numa_nodes {
-            nodes.push(NumaNode::new(id, numa_load_avg));
-        }
+        let mut nodes: Vec<NumaNode> = (0..num_numa_nodes)
+            .map(|id| NumaNode::new(id, numa_load_avg))
+            .collect();
 
         let dom_load_avg = total_load / dom_loads.len() as f64;
         for (dom_id, load) in dom_loads.iter().enumerate() {
-            let numa_id = self.dom_group.dom_numa_id(&dom_id).unwrap();
+            let numa_id = self
+                .dom_group
+                .dom_numa_id(&dom_id)
+                .ok_or_else(|| anyhow!("Failed to get NUMA ID for domain {}", dom_id))?;
 
             if numa_id >= num_numa_nodes {
                 bail!("NUMA ID {} exceeds maximum {}", numa_id, num_numa_nodes);
@@ -550,9 +553,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
             node.allocate_domain(dom_id, *load, dom_load_avg);
         }
 
-        for _ in 0..num_numa_nodes {
-            self.nodes.insert(nodes.pop().unwrap());
-        }
+        self.nodes = SortedVec::from_unsorted(nodes);
 
         Ok(())
     }
@@ -560,7 +561,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
     fn calculate_load_avgs(&mut self) -> Result<LoadLedger> {
         const NUM_BUCKETS: u64 = bpf_intf::consts_LB_LOAD_BUCKETS as u64;
         let now_mono = now_monotonic();
-        let load_half_life = self.skel.maps.rodata_data.load_half_life;
+        let load_half_life = self.skel.maps.rodata_data.as_ref().unwrap().load_half_life;
 
         let mut aggregator =
             LoadAggregator::new(self.dom_group.weight(), !self.lb_apply_weight.clone());
@@ -629,7 +630,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
 
         // Read active_tasks and update read_idx and gen.
         const MAX_TPTRS: u64 = bpf_intf::consts_MAX_DOM_ACTIVE_TPTRS as u64;
-        let dom_ctx = unsafe { &mut *self.skel.maps.bss_data.dom_ctxs[dom.id] };
+        let dom_ctx = unsafe { &mut *self.skel.maps.bss_data.as_mut().unwrap().dom_ctxs[dom.id] };
         let active_tasks = &mut dom_ctx.active_tasks;
 
         let (mut ridx, widx) = (active_tasks.read_idx, active_tasks.write_idx);
@@ -641,7 +642,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
         }
 
         // Read task_ctx and load.
-        let load_half_life = self.skel.maps.rodata_data.load_half_life;
+        let load_half_life = self.skel.maps.rodata_data.as_ref().unwrap().load_half_life;
         let now_mono = now_monotonic();
 
         for idx in ridx..widx {
