@@ -29,10 +29,14 @@ UEI_DEFINE(uei);
 
 const volatile u32 nr_cpu_ids;
 const volatile bool smt_enabled;
-const volatile bool prefer_same_cpu;
 const volatile u64 slice_ns;
 const volatile u64 tick_freq;
 
+/*
+ * CPUs sorted by their capacity in descendent order.
+ */
+const volatile u64 preferred_cpus[MAX_CPUS];
+
 /*
  * Scheduling statistics.
  */
@@ -41,7 +45,6 @@ volatile u64 nr_direct_dispatches, nr_timer_dispatches, nr_primary_dispatches;
 
 struct cpu_ctx {
 	struct bpf_timer timer;
-	u64 started_at;
 };
 
 struct {
@@ -251,14 +254,21 @@ void BPF_STRUCT_OPS(tickless_enqueue, struct task_struct *p, u64 enq_flags)
 	scx_bpf_dsq_insert_vtime(p, SHARED_DSQ, SCX_SLICE_INF, deadline, enq_flags);
 }
 
+/*
+ * Return true if @p can only run on a single CPU, false otherwise.
+ */
+static inline bool is_pcpu_task(const struct task_struct *p)
+{
+	return p->nr_cpus_allowed == 1 || is_migration_disabled(p);
+}
+
 /*
  * Try to consume a task from the shared queue and dispatch on a target
- * @cpu. If @same_cpu is true, try to consume a task that was previously
- * running on @cpu.
+ * @cpu.
  *
  * Return true if a task was dispatched, false otherwise.
  */
-static bool dispatch_cpu(s32 cpu, bool same_cpu, bool from_dispatch)
+static bool dispatch_cpu(s32 cpu, bool from_dispatch)
 {
 	struct task_struct *p;
 	bool dispatched = false;
@@ -274,14 +284,33 @@ static bool dispatch_cpu(s32 cpu, bool same_cpu, bool from_dispatch)
 		if (!p)
 			continue;
 
-		if (same_cpu && cpu != scx_bpf_task_cpu(p)) {
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) {
 			bpf_task_release(p);
 			continue;
 		}
 
-		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) {
+		/*
+		 * If there's an eligible task in the SHARED_DSQ that can
+		 * run on this CPU, clear the CPU's idle state and dispatch
+		 * the task. Otherwise, immediately give up and move to the
+		 * next CPU, since the goal is to avoid contention as much
+		 * as possible, so that tasks can keep using their infinite
+		 * time slice.
+		 *
+		 * However, if the first eligible task in the SHARED_DSQ is
+		 * a per-CPU task, dispatch it regardless of the CPU's idle
+		 * state, since contention is unavoidable in this case (the
+		 * task can only run on this CPU). Then the currently
+		 * running task will be made preemptible by the BPF timer,
+		 * changing its time slice from SCX_SLICE_INF to a finite
+		 * slice_ns.
+		 *
+		 * In this way we can still minimize CPU contention without
+		 * introducing starvation for per-CPU tasks.
+		 */
+		if (!scx_bpf_test_and_clear_cpu_idle(cpu) && !is_pcpu_task(p)) {
 			bpf_task_release(p);
-			continue;
+			break;
 		}
 
 		if (!__COMPAT_scx_bpf_dsq_move(BPF_FOR_EACH_ITER, p, SCX_DSQ_LOCAL_ON | cpu, 0)) {
@@ -314,24 +343,25 @@ static bool dispatch_cpu(s32 cpu, bool same_cpu, bool from_dispatch)
  */
 static bool dispatch_all_cpus(bool do_idle_smt, bool from_dispatch)
 {
-	const struct cpumask *idle_mask;
+	const struct cpumask *mask;
 	s32 i, cpu;
 	bool is_done = false;
 
 	if (!scx_bpf_dsq_nr_queued(SHARED_DSQ))
 		return true;
 
-	idle_mask = do_idle_smt ? scx_bpf_get_idle_smtmask() : scx_bpf_get_idle_cpumask();
+	if (!smt_enabled & do_idle_smt)
+		return false;
 
-	bpf_for(i, 0, bpf_cpumask_weight(idle_mask)) {
+	mask = do_idle_smt ? scx_bpf_get_idle_smtmask() : NULL;
+
+	bpf_for(i, 0, MIN(nr_cpu_ids, MAX_CPUS)) {
 		/*
-		 * Lock the first idle CPU available and attempt to dispatch a
-		 * task.
+		 * Try to lock the first idle CPU available and attempt to
+		 * dispatch a task.
 		 */
-		cpu = bpf_cpumask_first(idle_mask);
-		if (cpu >= nr_cpu_ids)
-			break;
-		if (!scx_bpf_test_and_clear_cpu_idle(cpu))
+		cpu = preferred_cpus[i];
+		if ((mask && !bpf_cpumask_test_cpu(cpu, mask)))
 			continue;
 
 		/*
@@ -348,11 +378,10 @@ static bool dispatch_all_cpus(bool do_idle_smt, bool from_dispatch)
 			continue;
 
 		/*
-		 * Try to dispatch a task that was using this CPU first, if
-		 * @prefer_same_cpu is enabled.
+		 * Consume a task from the shared DSQ and dispatch it to
+		 * the target CPU.
 		 */
-		if (!prefer_same_cpu || !dispatch_cpu(cpu, true, from_dispatch))
-			dispatch_cpu(cpu, false, from_dispatch);
+		dispatch_cpu(cpu, from_dispatch);
 
 		/*
 		 * Stop dispatching if all the pending tasks have been
@@ -364,41 +393,50 @@ static bool dispatch_all_cpus(bool do_idle_smt, bool from_dispatch)
 		}
 	}
 
-	scx_bpf_put_cpumask(idle_mask);
+	if (mask)
+		scx_bpf_put_cpumask(mask);
 
 	return is_done;
 }
 
 static int sched_timerfn(void *map, int *key, struct bpf_timer *timer)
 {
-	struct cpu_ctx *cctx;
-	u64 now = scx_bpf_now();
 	s32 cpu;
 
 	/*
 	 * Dispatch tasks on the available CPUs.
 	 */
-	if (!smt_enabled || !dispatch_all_cpus(true, false))
+	if (!dispatch_all_cpus(true, false))
 		dispatch_all_cpus(false, false);
 
 	/*
 	 * Check if we need to preempt the running tasks.
 	 */
 	bpf_for(cpu, 0, nr_cpu_ids) {
-		cctx = try_lookup_cpu_ctx(cpu);
-		if (!cctx)
-			continue;
+		struct task_struct *p;
 
-		if (!cctx->started_at ||
-		    time_before(now, cctx->started_at + slice_ns))
+		/*
+		 * Ignore CPU if idle task is running.
+		 */
+		p = scx_bpf_cpu_rq(cpu)->curr;
+		if (p->flags & PF_IDLE)
 			continue;
 
+		/*
+		 * Ignore CPUs without any task waiting.
+		 */
 		if (!scx_bpf_dsq_nr_queued(SCX_DSQ_LOCAL_ON | cpu) &&
 		    !scx_bpf_dsq_nr_queued(SHARED_DSQ))
 			continue;
 
-		scx_bpf_kick_cpu(cpu, SCX_KICK_PREEMPT);
-		__sync_fetch_and_add(&nr_preemptions, 1);
+		/*
+		 * Set a finite time slice to the running task, so that it
+		 * can be preempted.
+		 */
+		if (p->scx.slice == SCX_SLICE_INF) {
+			p->scx.slice = slice_ns;
+			__sync_fetch_and_add(&nr_preemptions, 1);
+		}
 	}
 
 	bpf_timer_start(timer, tick_interval_ns(), 0);
@@ -436,7 +474,7 @@ void BPF_STRUCT_OPS(tickless_dispatch, s32 cpu, struct task_struct *prev)
 		 * Try to bounce all the queued tasks to the available
 		 * tickless CPUs first.
 		 */
-		if (!smt_enabled || !dispatch_all_cpus(true, true))
+		if (!dispatch_all_cpus(true, true))
 			dispatch_all_cpus(false, true);
 	}
 
@@ -482,15 +520,8 @@ void BPF_STRUCT_OPS(tickless_runnable, struct task_struct *p, u64 enq_flags)
  */
 void BPF_STRUCT_OPS(tickless_running, struct task_struct *p)
 {
-	s32 cpu = scx_bpf_task_cpu(p);
-	struct cpu_ctx *cctx;
 	struct task_ctx *tctx;
 
-	cctx = try_lookup_cpu_ctx(cpu);
-	if (!cctx)
-		return;
-	cctx->started_at = scx_bpf_now() ? : 1;
-
 	tctx = try_lookup_task_ctx(p);
 	if (!tctx)
 		return;
@@ -518,16 +549,9 @@ void BPF_STRUCT_OPS(tickless_tick, struct task_struct *p)
  */
 void BPF_STRUCT_OPS(tickless_stopping, struct task_struct *p, bool runnable)
 {
-	s32 cpu = scx_bpf_task_cpu(p);
-	struct cpu_ctx *cctx;
 	struct task_ctx *tctx;
 	u64 slice;
 
-	cctx = try_lookup_cpu_ctx(cpu);
-	if (!cctx)
-		return;
-	cctx->started_at = 0;
-
 	tctx = try_lookup_task_ctx(p);
 	if (!tctx)
 		return;
