@@ -8,7 +8,7 @@
 #include <scx/ravg_impl.bpf.h>
 #include <lib/sdt_task.h>
 
-#include <scx/bpf_arena_common.h>
+#include <scx/bpf_arena_common.bpf.h>
 #include <scx/bpf_arena_spin_lock.h>
 
 #include <lib/cpumask.h>
@@ -17,7 +17,6 @@
 #include "types.h"
 #include "lb_domain.h"
 
-#include <scx/bpf_arena_common.h>
 #include <errno.h>
 #include <stdbool.h>
 #include <bpf/bpf_helpers.h>
