@@ -26,6 +26,8 @@
 
 const volatile bool mempolicy_affinity;
 
+#ifndef __TARGET_ARCH_arm64
+
 /*
  * Returns the dom mask for a node.
  */
@@ -90,6 +92,14 @@ static void task_set_preferred_mempolicy_dom_mask(struct task_struct *p,
 	return;
 }
 
+#else
+
+static void task_set_preferred_mempolicy_dom_mask(struct task_struct *p,
+                                                 task_ptr taskc)
+{}
+
+#endif
+
 static u32 task_pick_domain(task_ptr taskc, struct task_struct *p,
 			    const struct cpumask *cpumask)
 {
