@@ -137,6 +137,7 @@ use std::collections::VecDeque;
 use std::fmt;
 use std::sync::Arc;
 
+use anyhow::anyhow;
 use anyhow::bail;
 use anyhow::Result;
 use log::debug;
@@ -529,14 +530,16 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
         let num_numa_nodes = self.dom_group.nr_nodes();
         let numa_load_avg = total_load / num_numa_nodes as f64;
 
-        let mut nodes: Vec<NumaNode> = Vec::with_capacity(num_numa_nodes);
-        for id in 0..num_numa_nodes {
-            nodes.push(NumaNode::new(id, numa_load_avg));
-        }
+        let mut nodes: Vec<NumaNode> = (0..num_numa_nodes)
+            .map(|id| NumaNode::new(id, numa_load_avg))
+            .collect();
 
         let dom_load_avg = total_load / dom_loads.len() as f64;
         for (dom_id, load) in dom_loads.iter().enumerate() {
-            let numa_id = self.dom_group.dom_numa_id(&dom_id).unwrap();
+            let numa_id = self
+                .dom_group
+                .dom_numa_id(&dom_id)
+                .ok_or_else(|| anyhow!("Failed to get NUMA ID for domain {}", dom_id))?;
 
             if numa_id >= num_numa_nodes {
                 bail!("NUMA ID {} exceeds maximum {}", numa_id, num_numa_nodes);
@@ -546,9 +549,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
             node.allocate_domain(dom_id, *load, dom_load_avg);
         }
 
-        for _ in 0..num_numa_nodes {
-            self.nodes.insert(nodes.pop().unwrap());
-        }
+        self.nodes = SortedVec::from_unsorted(nodes);
 
         Ok(())
     }
@@ -556,7 +557,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
     fn calculate_load_avgs(&mut self) -> Result<LoadLedger> {
         const NUM_BUCKETS: u64 = bpf_intf::consts_LB_LOAD_BUCKETS as u64;
         let now_mono = now_monotonic();
-        let load_half_life = self.skel.maps.rodata_data.load_half_life;
+        let load_half_life = self.skel.maps.rodata_data.as_ref().unwrap().load_half_life;
 
         let mut aggregator =
             LoadAggregator::new(self.dom_group.weight(), !self.lb_apply_weight.clone());
@@ -627,7 +628,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
         const MAX_TPTRS: u64 = bpf_intf::consts_MAX_DOM_ACTIVE_TPTRS as u64;
 
         let types::topo_level(index) = types::topo_level::TOPO_LLC;
-        let ptr = self.skel.maps.bss_data.topo_nodes[index as usize][dom.id];
+        let ptr = self.skel.maps.bss_data.as_ref().unwrap().topo_nodes[index as usize][dom.id];
         let dom_ctx = unsafe { std::mem::transmute::<u64, &mut types::dom_ctx>(ptr) };
         let active_tasks = &mut dom_ctx.active_tasks;
 
@@ -640,7 +641,7 @@ impl<'a, 'b> LoadBalancer<'a, 'b> {
         }
 
         // Read task_ctx and load.
-        let load_half_life = self.skel.maps.rodata_data.load_half_life;
+        let load_half_life = self.skel.maps.rodata_data.as_ref().unwrap().load_half_life;
         let now_mono = now_monotonic();
 
         for idx in ridx..widx {
