@@ -21,29 +21,29 @@
 
 #define CLOCK_BOOTTIME 7
 
-char _license[] SEC("license") = "GPL";
+char		    _license[] SEC("license")	     = "GPL";
 
-const volatile u64 long_tail_tracing_min_latency_ns = 0;
-const volatile bool layered = false;
+const volatile u64  long_tail_tracing_min_latency_ns = 0;
+const volatile bool layered			     = false;
 
-u64 trace_duration_ns = 1000000000;
-u64 trace_warmup_ns = 500000000;
-u64 last_trace_end_time = 0;
+u64		    trace_duration_ns		     = 1000000000;
+u64		    trace_warmup_ns		     = 500000000;
+u64		    last_trace_end_time		     = 0;
 
 // dummy for generating types
-struct bpf_event _event = {0};
+struct bpf_event _event		   = { 0 };
 
-bool enable_bpf_events = true;
+bool		 enable_bpf_events = true;
 
-enum mode mode = MODE_NORMAL;
-u32 sample_rate = 128;
-u32 last_sample_rate;
+enum mode	 mode		   = MODE_NORMAL;
+u32		 sample_rate	   = 128;
+u32		 last_sample_rate;
 
-const int zero_int = 0;
+const int	 zero_int = 0;
 
 struct timer_wrapper {
-	struct bpf_timer	timer;
-	int			key;
+	struct bpf_timer timer;
+	int		 key;
 };
 
 enum scxtop_timer_callbacks {
@@ -68,7 +68,7 @@ struct {
 	__uint(key_size, sizeof(u32));
 	__uint(value_size, sizeof(u64));
 	__uint(max_entries, NR_SCXTOP_STATS);
-} stats SEC(".maps");
+} stats			    SEC(".maps");
 
 static __always_inline void stat_inc(u32 idx)
 {
@@ -77,11 +77,12 @@ static __always_inline void stat_inc(u32 idx)
 		(*cnt_p)++;
 }
 
-static __always_inline struct bpf_event* try_reserve_event()
+static __always_inline struct bpf_event *try_reserve_event()
 {
 	struct bpf_event *event = NULL;
 
-	if (!(event = bpf_ringbuf_reserve(&events, sizeof(struct bpf_event), 0)))
+	if (!(event = bpf_ringbuf_reserve(&events, sizeof(struct bpf_event),
+					  0)))
 		stat_inc(STAT_DROPPED_EVENTS);
 
 	return event;
@@ -116,7 +117,7 @@ struct {
 	__uint(map_flags, BPF_F_NO_PREALLOC);
 	__type(key, int);
 	__type(value, void *);
-} task_ctxs SEC(".maps");
+} task_ctxs		    SEC(".maps");
 
 static __always_inline u32 *try_lookup_layered_task_ctx(struct task_struct *p)
 {
@@ -131,8 +132,8 @@ struct {
 } long_tail_entries SEC(".maps");
 
 struct __softirq_event {
-	u32		pid;
-	u64		start_ts;
+	u32 pid;
+	u64 start_ts;
 };
 
 struct {
@@ -140,7 +141,7 @@ struct {
 	__type(key, int);
 	__type(value, struct __softirq_event);
 	__uint(max_entries, 1);
-} softirq_events SEC(".maps");
+} softirq_events	   SEC(".maps");
 
 static __always_inline u64 t_to_tptr(struct task_struct *p)
 {
@@ -157,7 +158,7 @@ static __always_inline u64 t_to_tptr(struct task_struct *p)
 static struct task_ctx *try_lookup_task_ctx(struct task_struct *p)
 {
 	struct task_ctx *tctx;
-	u64 tptr;
+	u64		 tptr;
 
 	if (!p)
 		return NULL;
@@ -169,12 +170,14 @@ static struct task_ctx *try_lookup_task_ctx(struct task_struct *p)
 	tctx = bpf_map_lookup_elem(&task_data, &tptr);
 	if (!tctx) {
 		struct task_ctx new_tctx;
-		new_tctx.dsq_id = SCX_DSQ_INVALID;
-		new_tctx.dsq_vtime = 0;
-		new_tctx.slice_ns = 0;
-		new_tctx.last_run_ns = 0;
+		new_tctx.dsq_id		 = SCX_DSQ_INVALID;
+		new_tctx.dsq_vtime	 = 0;
+		new_tctx.slice_ns	 = 0;
+		new_tctx.last_run_ns	 = 0;
 		new_tctx.dsq_insert_time = 0;
-		new_tctx.wakeup_ts = 0;
+		new_tctx.wakeup_ts	 = 0;
+		new_tctx.last_waker_pid	 = 0;
+		__builtin_memset(new_tctx.last_waker_comm, 0, MAX_COMM);
 
 		if (!bpf_map_update_elem(&task_data, &tptr, &new_tctx, BPF_ANY))
 			return NULL;
@@ -184,7 +187,8 @@ static struct task_ctx *try_lookup_task_ctx(struct task_struct *p)
 	return tctx;
 }
 
-static int update_task_ctx(struct task_struct *p, u64 dsq, u64 vtime, u64 slice_ns)
+static int update_task_ctx(struct task_struct *p, u64 dsq, u64 vtime,
+			   u64 slice_ns)
 {
 	if (!enable_bpf_events)
 		return 0;
@@ -195,9 +199,9 @@ static int update_task_ctx(struct task_struct *p, u64 dsq, u64 vtime, u64 slice_
 		return -ENOENT;
 
 	tctx->dsq_insert_time = bpf_ktime_get_ns();
-	tctx->dsq_id = dsq;
-	tctx->dsq_vtime = vtime;
-	tctx->slice_ns = slice_ns;
+	tctx->dsq_id	      = dsq;
+	tctx->dsq_vtime	      = vtime;
+	tctx->slice_ns	      = slice_ns;
 
 	return 0;
 }
@@ -213,9 +217,9 @@ int generic_kprobe(struct pt_regs *ctx)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = KPROBE;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
+	event->type		= KPROBE;
+	event->cpu		= bpf_get_smp_processor_id();
+	event->ts		= bpf_ktime_get_ns();
 	event->event.kprobe.pid = bpf_get_current_pid_tgid() & 0xffffffff;
 	event->event.kprobe.instruction_pointer = bpf_get_func_ip(ctx);
 
@@ -269,8 +273,8 @@ int BPF_KPROBE(on_sched_cpu_perf, s32 cpu, u32 perf)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = CPU_PERF_SET;
-	event->cpu = cpu;
+	event->type	       = CPU_PERF_SET;
+	event->cpu	       = cpu;
 	event->event.perf.perf = perf;
 	bpf_ringbuf_submit(event, 0);
 
@@ -278,13 +282,15 @@ int BPF_KPROBE(on_sched_cpu_perf, s32 cpu, u32 perf)
 }
 
 SEC("kprobe/scx_bpf_dsq_insert_vtime")
-int BPF_KPROBE(scx_insert_vtime, struct task_struct *p, u64 dsq, u64 slice_ns, u64 vtime)
+int BPF_KPROBE(scx_insert_vtime, struct task_struct *p, u64 dsq, u64 slice_ns,
+	       u64 vtime)
 {
 	return update_task_ctx(p, dsq, vtime, slice_ns);
 }
 
 SEC("kprobe/scx_bpf_dispatch_vtime")
-int BPF_KPROBE(scx_dispatch_vtime, struct task_struct *p, u64 dsq, u64 slice_ns, u64 vtime)
+int BPF_KPROBE(scx_dispatch_vtime, struct task_struct *p, u64 dsq, u64 slice_ns,
+	       u64 vtime)
 {
 	return update_task_ctx(p, dsq, vtime, slice_ns);
 }
@@ -300,8 +306,8 @@ static int on_insert(struct task_struct *p, u64 dsq)
 		return -ENOENT;
 
 	tctx->dsq_insert_time = bpf_ktime_get_ns();
-	tctx->dsq_id = dsq;
-	tctx->dsq_vtime = 0;
+	tctx->dsq_id	      = dsq;
+	tctx->dsq_vtime	      = 0;
 
 	return 0;
 }
@@ -329,22 +335,22 @@ static int on_dsq_move(struct task_struct *p, u64 dsq)
 		return -ENOENT;
 
 	tctx->dsq_insert_time = bpf_ktime_get_ns();
-	tctx->dsq_id = dsq;
-	tctx->dsq_vtime = 0;
+	tctx->dsq_id	      = dsq;
+	tctx->dsq_vtime	      = 0;
 
 	return 0;
 }
 
 SEC("kprobe/scx_bpf_dsq_move")
 int BPF_KPROBE(scx_dsq_move, struct bpf_iter_scx_dsq *it__iter,
-               struct task_struct *p, u64 dsq_id, u64 enq_flags)
+	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
 	return on_dsq_move(p, dsq_id);
 }
 
 SEC("kprobe/scx_bpf_dispatch_from_dsq")
 int BPF_KPROBE(scx_dispatch_from_dsq, struct bpf_iter_scx_dsq *it__iter,
-               struct task_struct *p, u64 dsq_id, u64 enq_flags)
+	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
 	return on_dsq_move(p, dsq_id);
 }
@@ -360,7 +366,7 @@ static int on_dsq_move_vtime(struct task_struct *p, u64 dsq)
 		return -ENOENT;
 
 	tctx->dsq_insert_time = bpf_ktime_get_ns();
-	tctx->dsq_id = dsq;
+	tctx->dsq_id	      = dsq;
 	bpf_core_read(&tctx->dsq_vtime, sizeof(u64), &p->scx.dsq_vtime);
 
 	return 0;
@@ -368,14 +374,14 @@ static int on_dsq_move_vtime(struct task_struct *p, u64 dsq)
 
 SEC("kprobe/scx_bpf_dsq_move_vtime")
 int BPF_KPROBE(scx_dsq_move_vtime, struct bpf_iter_scx_dsq *it__iter,
-               struct task_struct *p, u64 dsq_id, u64 enq_flags)
+	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
 	return on_dsq_move_vtime(p, dsq_id);
 }
 
 SEC("kprobe/scx_bpf_dispatch_vtime_from_dsq")
 int BPF_KPROBE(scx_dispatch_vtime_from_dsq, struct bpf_iter_scx_dsq *it__iter,
-               struct task_struct *p, u64 dsq_id, u64 enq_flags)
+	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
 	return on_dsq_move_vtime(p, dsq_id);
 }
@@ -396,7 +402,8 @@ static int on_move_set_slice(struct task_struct *p, u64 slice)
 }
 
 SEC("kprobe/scx_bpf_dsq_move_set_slice")
-int BPF_KPROBE(scx_dsq_move_set_slice, struct bpf_iter_scx_dsq *it__iter, u64 slice)
+int BPF_KPROBE(scx_dsq_move_set_slice, struct bpf_iter_scx_dsq *it__iter,
+	       u64 slice)
 {
 	// TODO: figure out how to return task from iterator without consuming.
 	return on_move_set_slice(NULL, slice);
@@ -404,7 +411,7 @@ int BPF_KPROBE(scx_dsq_move_set_slice, struct bpf_iter_scx_dsq *it__iter, u64 sl
 
 SEC("kprobe/scx_bpf_dispatch_from_dsq_set_slice")
 int BPF_KPROBE(scx_dispatch_from_dsq_set_slice,
-               struct bpf_iter_scx_dsq *it__iter, u64 slice)
+	       struct bpf_iter_scx_dsq *it__iter, u64 slice)
 {
 	// TODO: figure out how to return task from iterator without consuming.
 	return on_move_set_slice(NULL, slice);
@@ -427,7 +434,7 @@ static int on_move_set_vtime(struct task_struct *p, u64 vtime)
 
 SEC("kprobe/scx_bpf_dsq_move_set_vtime")
 int BPF_KPROBE(scx_dsq_move_set_vtime, struct bpf_iter_scx_dsq *it__iter,
-               u64 vtime)
+	       u64 vtime)
 {
 	// TODO: figure out how to return task from iterator without consuming.
 	return on_move_set_vtime(NULL, vtime);
@@ -435,7 +442,7 @@ int BPF_KPROBE(scx_dsq_move_set_vtime, struct bpf_iter_scx_dsq *it__iter,
 
 SEC("kprobe/scx_bpf_dispatch_from_dsq_set_vtime")
 int BPF_KPROBE(scx_dispatch_from_dsq_set_vtime,
-               struct bpf_iter_scx_dsq *it__iter, u64 vtime)
+	       struct bpf_iter_scx_dsq *it__iter, u64 vtime)
 {
 	// TODO: figure out how to return task from iterator without consuming.
 	return on_move_set_vtime(NULL, vtime);
@@ -449,7 +456,8 @@ static void record_real_comm(u8 *comm, struct task_struct *task)
 		 * stick with the measly 16 characters of the comm field to
 		 * keep things simple.
 		 */
-		struct kthread *k = bpf_core_cast(task->worker_private, struct kthread);
+		struct kthread *k =
+			bpf_core_cast(task->worker_private, struct kthread);
 		struct worker *worker = bpf_core_cast(k->data, struct worker);
 		bpf_probe_read_kernel_str(comm, MAX_COMM, worker->desc);
 	} else {
@@ -459,30 +467,48 @@ static void record_real_comm(u8 *comm, struct task_struct *task)
 
 static __always_inline int __on_sched_wakeup(struct task_struct *p)
 {
-	struct task_ctx *tctx;
-	struct bpf_event *event;
+	struct task_ctx	   *tctx;
+	struct bpf_event   *event;
+	struct task_struct *waker;
 
 	if (!enable_bpf_events || !p || !should_sample())
 		return 0;
 
 	u64 now = bpf_ktime_get_ns();
-	tctx = try_lookup_task_ctx(p);
+	tctx	= try_lookup_task_ctx(p);
+
+	// Get the current task (the waker)
+	waker = (struct task_struct *)bpf_get_current_task_btf();
+
+	// Record waker information in the target task's context
+	if (tctx && waker) {
+		tctx->wakeup_ts	     = now;
+		tctx->last_waker_pid = waker->pid;
+		record_real_comm(tctx->last_waker_comm, waker);
+	} else if (tctx) {
+		tctx->wakeup_ts = now;
+	}
 
 	if (!(event = try_reserve_event()))
 		return 0;
 
-	event->type = SCHED_WAKEUP;
-
-	if (tctx)
-		tctx->wakeup_ts = now;
-
-	event->ts = now;
-	event->cpu = bpf_get_smp_processor_id();
-	event->event.wakeup.pid = p->pid;
+	event->type		 = SCHED_WAKEUP;
+	event->ts		 = now;
+	event->cpu		 = bpf_get_smp_processor_id();
+	event->event.wakeup.pid	 = p->pid;
 	event->event.wakeup.tgid = p->tgid;
 	event->event.wakeup.prio = (int)p->prio;
 	record_real_comm(event->event.wakeup.comm, p);
 
+	// Include waker information in the event
+	if (waker) {
+		event->event.wakeup.waker_pid = waker->pid;
+		record_real_comm(event->event.wakeup.waker_comm, waker);
+	} else {
+		event->event.wakeup.waker_pid = 0;
+		__builtin_memset(event->event.wakeup.waker_comm, 0, MAX_COMM);
+	}
+
 	bpf_ringbuf_submit(event, 0);
 
 	return 0;
@@ -503,31 +529,54 @@ int BPF_PROG(on_sched_wakeup_new, struct task_struct *p)
 SEC("tp_btf/sched_waking")
 int BPF_PROG(on_sched_waking, struct task_struct *p)
 {
-	struct bpf_event *event;
+	struct task_ctx	   *tctx;
+	struct bpf_event   *event;
+	struct task_struct *waker;
 
 	if (!enable_bpf_events || !should_sample())
 		return 0;
 
+	u64 now = bpf_ktime_get_ns();
+	tctx	= try_lookup_task_ctx(p);
+
+	// Get the current task (the waker)
+	waker = (struct task_struct *)bpf_get_current_task_btf();
+
+	// Record waker information in the target task's context
+	if (tctx && waker) {
+		tctx->last_waker_pid = waker->pid;
+		record_real_comm(tctx->last_waker_comm, waker);
+	}
+
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = SCHED_WAKING;
-	event->ts = bpf_ktime_get_ns();
-	event->cpu = bpf_get_smp_processor_id();
-	event->event.wakeup.pid = p->pid;
+	event->type		 = SCHED_WAKING;
+	event->ts		 = now;
+	event->cpu		 = bpf_get_smp_processor_id();
+	event->event.wakeup.pid	 = p->pid;
 	event->event.wakeup.tgid = p->tgid;
 	event->event.wakeup.prio = (int)p->prio;
 	record_real_comm(event->event.wakeup.comm, p);
 
+	// Include waker information in the event
+	if (waker) {
+		event->event.wakeup.waker_pid = waker->pid;
+		record_real_comm(event->event.wakeup.waker_comm, waker);
+	} else {
+		event->event.wakeup.waker_pid = 0;
+		__builtin_memset(event->event.wakeup.waker_comm, 0, MAX_COMM);
+	}
+
 	bpf_ringbuf_submit(event, 0);
 	return 0;
 }
 
 SEC("tp_btf/sched_switch")
 int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
-             struct task_struct *next, u64 prev_state)
+	     struct task_struct *next, u64 prev_state)
 {
-	struct task_ctx *next_tctx, *prev_tctx;
+	struct task_ctx	 *next_tctx, *prev_tctx;
 	struct bpf_event *event;
 
 	prev_tctx = try_lookup_task_ctx(prev);
@@ -539,31 +588,33 @@ int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
 		if (!(event = try_reserve_event()))
 			return -ENOMEM;
 
-		u64 now = bpf_ktime_get_ns();
+		u64 now	    = bpf_ktime_get_ns();
 		event->type = SCHED_SWITCH;
-		event->cpu = bpf_get_smp_processor_id();
-		event->ts = now;
+		event->cpu  = bpf_get_smp_processor_id();
+		event->ts   = now;
 
 		if (layered && (lctx = try_lookup_layered_task_ctx(prev)))
-			event->event.sched_switch.prev_layer_id = lctx[LAYER_ID_INDEX];
+			event->event.sched_switch.prev_layer_id =
+				lctx[LAYER_ID_INDEX];
 		else
 			event->event.sched_switch.prev_layer_id = -1;
 
-		event->event.sched_switch.preempt = preempt;
-		event->event.sched_switch.prev_pid = prev->pid;
-		event->event.sched_switch.prev_tgid = prev->tgid;
-		event->event.sched_switch.prev_prio = (int)prev->prio;
+		event->event.sched_switch.preempt    = preempt;
+		event->event.sched_switch.prev_pid   = prev->pid;
+		event->event.sched_switch.prev_tgid  = prev->tgid;
+		event->event.sched_switch.prev_prio  = (int)prev->prio;
 		event->event.sched_switch.prev_state = prev_state;
-		event->event.sched_switch.prev_used_slice_ns = now - prev_tctx->last_run_ns;
-		event->event.sched_switch.prev_dsq_id = prev_tctx->dsq_id;
+		event->event.sched_switch.prev_used_slice_ns =
+			now - prev_tctx->last_run_ns;
+		event->event.sched_switch.prev_dsq_id	= prev_tctx->dsq_id;
 		event->event.sched_switch.prev_slice_ns = prev_tctx->slice_ns;
 		record_real_comm(event->event.sched_switch.prev_comm, prev);
 
-		prev_tctx->dsq_id = SCX_DSQ_INVALID;
-		prev_tctx->dsq_vtime = 0;
-		prev_tctx->wakeup_ts = 0;
+		prev_tctx->dsq_id	   = SCX_DSQ_INVALID;
+		prev_tctx->dsq_vtime	   = 0;
+		prev_tctx->wakeup_ts	   = 0;
 		prev_tctx->dsq_insert_time = 0;
-		prev_tctx->last_run_ns = 0;
+		prev_tctx->last_run_ns	   = 0;
 
 		/*
 		 * Tracking vtime **and** the dsq a task was inserted to is kind of
@@ -575,23 +626,29 @@ int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
 		 * when tasks are moved from iterators.
 		*/
 		if (next) {
-			if (layered && (lctx = try_lookup_layered_task_ctx(next)))
-				event->event.sched_switch.next_layer_id = lctx[LAYER_ID_INDEX];
+			if (layered &&
+			    (lctx = try_lookup_layered_task_ctx(next)))
+				event->event.sched_switch.next_layer_id =
+					lctx[LAYER_ID_INDEX];
 			else
 				event->event.sched_switch.next_layer_id = -1;
 
 			next_tctx = try_lookup_task_ctx(next);
-			event->event.sched_switch.next_pid = next->pid;
+			event->event.sched_switch.next_pid  = next->pid;
 			event->event.sched_switch.next_tgid = next->tgid;
 			event->event.sched_switch.next_prio = (int)next->prio;
-			record_real_comm(event->event.sched_switch.next_comm, next);
+			record_real_comm(event->event.sched_switch.next_comm,
+					 next);
 
 			if (next_tctx && next_tctx->dsq_insert_time > 0) {
 				event->event.sched_switch.next_dsq_lat_us =
-					(now - next_tctx->dsq_insert_time) / 1000;
-				event->event.sched_switch.next_dsq_id = next_tctx->dsq_id;
+					(now - next_tctx->dsq_insert_time) /
+					1000;
+				event->event.sched_switch.next_dsq_id =
+					next_tctx->dsq_id;
 				event->event.sched_switch.next_dsq_nr =
-					scx_bpf_dsq_nr_queued(next_tctx->dsq_id);
+					scx_bpf_dsq_nr_queued(
+						next_tctx->dsq_id);
 				/*
 				 * XXX: if a task gets moved to another dsq and the vtime is updated
 				 * then vtime should be read off the sched_ext_entity. To properly
@@ -599,17 +656,19 @@ int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
 				 * updated the tctx needs to be updated.
 				*/
 				// bpf_core_read(&event->dsq_vtime, sizeof(u64), &p->scx.dsq_vtime);
-				event->event.sched_switch.next_dsq_vtime = next_tctx->dsq_vtime;
+				event->event.sched_switch.next_dsq_vtime =
+					next_tctx->dsq_vtime;
 			} else {
-				event->event.sched_switch.next_dsq_id = SCX_DSQ_INVALID;
+				event->event.sched_switch.next_dsq_id =
+					SCX_DSQ_INVALID;
 				event->event.sched_switch.next_dsq_lat_us = 0;
-				event->event.sched_switch.next_dsq_nr = 0;
-				event->event.sched_switch.next_dsq_vtime = 0;
+				event->event.sched_switch.next_dsq_nr	  = 0;
+				event->event.sched_switch.next_dsq_vtime  = 0;
 			}
 		} else {
 			event->event.sched_switch.next_dsq_lat_us = 0;
-			event->event.sched_switch.next_pid = 0;
-			event->event.sched_switch.next_tgid = 0;
+			event->event.sched_switch.next_pid	  = 0;
+			event->event.sched_switch.next_tgid	  = 0;
 		}
 
 		bpf_ringbuf_submit(event, 0);
@@ -620,10 +679,10 @@ int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
 		return 0;
 
 	if ((next_tctx = try_lookup_task_ctx(next))) {
-		next_tctx->last_run_ns = bpf_ktime_get_ns();
-		next_tctx->dsq_vtime = 0;
+		next_tctx->last_run_ns	   = bpf_ktime_get_ns();
+		next_tctx->dsq_vtime	   = 0;
 		next_tctx->dsq_insert_time = 0;
-		next_tctx->wakeup_ts = 0;
+		next_tctx->wakeup_ts	   = 0;
 	}
 
 	return 0;
@@ -640,12 +699,12 @@ int BPF_PROG(on_sched_migrate_task, struct task_struct *p, int dest_cpu)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = SCHED_MIGRATE;
-	event->ts = bpf_ktime_get_ns();
-	event->cpu = bpf_get_smp_processor_id();
-	event->event.migrate.pid = p->pid;
+	event->type		      = SCHED_MIGRATE;
+	event->ts		      = bpf_ktime_get_ns();
+	event->cpu		      = bpf_get_smp_processor_id();
+	event->event.migrate.pid      = p->pid;
 	event->event.migrate.dest_cpu = dest_cpu;
-	event->event.migrate.prio = (int)p->prio;
+	event->event.migrate.prio     = (int)p->prio;
 
 	bpf_ringbuf_submit(event, 0);
 
@@ -664,8 +723,8 @@ int BPF_PROG(on_sched_hang, struct task_struct *p)
 		return -ENOMEM;
 
 	event->type = SCHED_HANG;
-	event->ts = bpf_ktime_get_ns();
-	event->cpu = bpf_get_smp_processor_id();
+	event->ts   = bpf_ktime_get_ns();
+	event->cpu  = bpf_get_smp_processor_id();
 	record_real_comm(event->event.hang.comm, p);
 	event->event.hang.pid = p->pid;
 
@@ -699,13 +758,13 @@ int BPF_PROG(on_softirq_entry, unsigned int nr)
 SEC("tp_btf/softirq_exit")
 int BPF_PROG(on_softirq_exit, unsigned int nr)
 {
-	struct bpf_event *event;
+	struct bpf_event       *event;
 	struct __softirq_event *softirq_event;
 
 	if (!enable_bpf_events || !should_sample())
 		return 0;
 
-	u64 exit_ts = bpf_ktime_get_ns();
+	u64 exit_ts   = bpf_ktime_get_ns();
 
 	softirq_event = bpf_map_lookup_elem(&softirq_events, &zero_int);
 	if (!softirq_event)
@@ -716,12 +775,12 @@ int BPF_PROG(on_softirq_exit, unsigned int nr)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = SOFTIRQ;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = exit_ts;
-	event->event.softirq.pid = softirq_event->pid;
-	event->event.softirq.entry_ts = softirq_event->start_ts;
-	event->event.softirq.exit_ts = exit_ts;
+	event->type			= SOFTIRQ;
+	event->cpu			= bpf_get_smp_processor_id();
+	event->ts			= exit_ts;
+	event->event.softirq.pid	= softirq_event->pid;
+	event->event.softirq.entry_ts	= softirq_event->start_ts;
+	event->event.softirq.exit_ts	= exit_ts;
 	event->event.softirq.softirq_nr = nr;
 
 	bpf_ringbuf_submit(event, 0);
@@ -730,17 +789,18 @@ int BPF_PROG(on_softirq_exit, unsigned int nr)
 }
 
 static int stop_trace_timer_callback(void *map, int key,
-                                     struct timer_wrapper *timerw)
+				     struct timer_wrapper *timerw)
 {
 	struct bpf_event *event;
-	u64 end = mode == MODE_TRACING ? bpf_ktime_get_ns() : last_trace_end_time;
+	u64		  end = mode == MODE_TRACING ? bpf_ktime_get_ns() :
+						       last_trace_end_time;
 
-	sample_rate = last_sample_rate;
+	sample_rate	      = last_sample_rate;
 
 	if ((event = try_reserve_event())) {
-		mode = MODE_NORMAL;
+		mode	    = MODE_NORMAL;
 
-		event->ts = end;
+		event->ts   = end;
 		event->type = TRACE_STOPPED;
 
 		bpf_ringbuf_submit(event, 0);
@@ -750,7 +810,7 @@ static int stop_trace_timer_callback(void *map, int key,
 	// Failed to get event. We've already slowed down the sample rate which
 	// will reduce the amount of events userspace needs to handle. Log when
 	// the trace actually ended and retry in 5ms.
-	mode = MODE_TRACE_STOPPING;
+	mode		    = MODE_TRACE_STOPPING;
 	last_trace_end_time = end;
 
 	bpf_timer_start(&timerw->timer, 5000000, 0);
@@ -758,21 +818,23 @@ static int stop_trace_timer_callback(void *map, int key,
 }
 
 static __always_inline int start_trace_real(bool schedule_stop,
-                                            bool start_immediately)
+					    bool start_immediately)
 {
-	static const enum scxtop_timer_callbacks stop_trace_key = TIMER_STOP_TRACE;
+	static const enum scxtop_timer_callbacks stop_trace_key =
+		TIMER_STOP_TRACE;
 
 	u64 duration_ns = trace_duration_ns;
 	if (!start_immediately)
 		duration_ns += trace_warmup_ns;
 
 	// do not restart a started trace. this may be relaxed in future.
-	enum mode last_mode = __sync_val_compare_and_swap(&mode, MODE_NORMAL, MODE_TRACING);
+	enum mode last_mode =
+		__sync_val_compare_and_swap(&mode, MODE_NORMAL, MODE_TRACING);
 	if (last_mode == MODE_TRACING)
 		return 0;
 
 	struct timer_wrapper *timerw;
-	struct bpf_event *event;
+	struct bpf_event     *event;
 
 	// replicate the actions of userspace starting a trace so it starts
 	// immediately, but such that events will come after our ringbuffer
@@ -782,7 +844,7 @@ static __always_inline int start_trace_real(bool schedule_stop,
 	// always and activate them with a global, which could be expensive. for
 	// now let them start late.
 	last_sample_rate = sample_rate;
-	sample_rate = 1;
+	sample_rate	 = 1;
 
 	// inform userspace that following events are in trace mode
 	if (!(event = try_reserve_event()))
@@ -796,10 +858,10 @@ static __always_inline int start_trace_real(bool schedule_stop,
 			goto error_no_timer;
 	}
 
-	event->ts = bpf_ktime_get_ns();
-	event->type = TRACE_STARTED;
+	event->ts			     = bpf_ktime_get_ns();
+	event->type			     = TRACE_STARTED;
 	event->event.trace.start_immediately = start_immediately;
-	event->event.trace.stop_scheduled = schedule_stop;
+	event->event.trace.stop_scheduled    = schedule_stop;
 
 	bpf_ringbuf_submit(event, 0);
 	return 0;
@@ -819,7 +881,8 @@ error_no_event:
 SEC("syscall")
 int BPF_PROG(start_trace)
 {
-	start_trace_real(true /* schedule_stop */, false /* start_immediately */);
+	start_trace_real(true /* schedule_stop */,
+			 false /* start_immediately */);
 	return 0;
 }
 
@@ -827,7 +890,7 @@ SEC("uprobe")
 int BPF_UPROBE(long_tail_tracker_entry)
 {
 	u64 pidtgid = bpf_get_current_pid_tgid();
-	u64 ts = bpf_ktime_get_ns();
+	u64 ts	    = bpf_ktime_get_ns();
 	bpf_map_update_elem(&long_tail_entries, &pidtgid, &ts, BPF_ANY);
 	return 0;
 }
@@ -836,9 +899,9 @@ SEC("uretprobe")
 int BPF_URETPROBE(long_tail_tracker_exit)
 {
 	u64 *entry_time;
-	u64 now = bpf_ktime_get_ns();
+	u64  now     = bpf_ktime_get_ns();
 
-	u64 pidtgid = bpf_get_current_pid_tgid();
+	u64  pidtgid = bpf_get_current_pid_tgid();
 	if (!(entry_time = bpf_map_lookup_elem(&long_tail_entries, &pidtgid)))
 		return -ENOENT;
 
@@ -847,23 +910,27 @@ int BPF_URETPROBE(long_tail_tracker_exit)
 
 	// we can't start the trace fully from the bpf side directly here
 	// because we need to schedule the timer that terminates the trace, and:
-	//	tracing progs cannot use bpf_timer yet 
+	//	tracing progs cannot use bpf_timer yet
 	// instead start the trace but include in the message to userspace the
 	// fact we haven't scheduled the stop, and have userspace call back
 	// into a "syscall" type program which can schedule the stop. userspace
 	// can compute the absolute stop time to make this less racy.
-	return start_trace_real(false /* schedule_stop */, true /* start_immediately */);
+	return start_trace_real(false /* schedule_stop */,
+				true /* start_immediately */);
 }
 
 SEC("syscall")
 int schedule_stop_trace(struct schedule_stop_trace_args *args)
 {
-	static const enum scxtop_timer_callbacks stop_trace_key = TIMER_STOP_TRACE;
+	static const enum scxtop_timer_callbacks stop_trace_key =
+		TIMER_STOP_TRACE;
 
-	struct timer_wrapper *timerw = bpf_map_lookup_elem(&timers, &stop_trace_key);
+	struct timer_wrapper *timerw =
+		bpf_map_lookup_elem(&timers, &stop_trace_key);
 	if (!timerw)
 		return -ENOENT;
-	if (bpf_timer_start(&timerw->timer, args->stop_timestamp, BPF_F_TIMER_ABS) < 0)
+	if (bpf_timer_start(&timerw->timer, args->stop_timestamp,
+			    BPF_F_TIMER_ABS) < 0)
 		return -ENOENT;
 
 	return 0;
@@ -872,7 +939,7 @@ int schedule_stop_trace(struct schedule_stop_trace_args *args)
 SEC("tp_btf/ipi_send_cpu")
 int BPF_PROG(on_ipi_send_cpu, u32 cpu, void *callsite, void *callback)
 {
-	struct bpf_event *event;
+	struct bpf_event   *event;
 	struct task_struct *p;
 
 	if (!enable_bpf_events || !should_sample())
@@ -881,9 +948,9 @@ int BPF_PROG(on_ipi_send_cpu, u32 cpu, void *callsite, void *callback)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = IPI;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
+	event->type		    = IPI;
+	event->cpu		    = bpf_get_smp_processor_id();
+	event->ts		    = bpf_ktime_get_ns();
 	event->event.ipi.target_cpu = cpu;
 
 	p = (struct task_struct *)bpf_get_current_task();
@@ -908,10 +975,10 @@ int BPF_PROG(on_sched_exit, struct task_struct *task)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = EXIT;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
-	event->event.exit.pid = BPF_CORE_READ(task, pid);
+	event->type	       = EXIT;
+	event->cpu	       = bpf_get_smp_processor_id();
+	event->ts	       = bpf_ktime_get_ns();
+	event->event.exit.pid  = BPF_CORE_READ(task, pid);
 	event->event.exit.tgid = BPF_CORE_READ(task, tgid);
 	event->event.exit.prio = BPF_CORE_READ(task, prio);
 	record_real_comm(event->event.exit.comm, task);
@@ -923,24 +990,24 @@ int BPF_PROG(on_sched_exit, struct task_struct *task)
 
 SEC("tp_btf/sched_process_fork")
 int BPF_PROG(on_sched_fork, struct task_struct *parent,
-             struct task_struct *child)
+	     struct task_struct *child)
 {
 	struct bpf_event *event;
-	u32 *lctx;
+	u32		 *lctx;
 
 	if (!enable_bpf_events)
-	return 0;
+		return 0;
 
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = FORK;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
-	event->event.fork.parent_pid = BPF_CORE_READ(parent, pid);
+	event->type		      = FORK;
+	event->cpu		      = bpf_get_smp_processor_id();
+	event->ts		      = bpf_ktime_get_ns();
+	event->event.fork.parent_pid  = BPF_CORE_READ(parent, pid);
 	event->event.fork.parent_tgid = BPF_CORE_READ(parent, tgid);
-	event->event.fork.child_pid = BPF_CORE_READ(child, pid);
-	event->event.fork.child_tgid = BPF_CORE_READ(child, tgid);
+	event->event.fork.child_pid   = BPF_CORE_READ(child, pid);
+	event->event.fork.child_tgid  = BPF_CORE_READ(child, tgid);
 	record_real_comm(event->event.fork.parent_comm, parent);
 	record_real_comm(event->event.fork.child_comm, child);
 
@@ -960,11 +1027,11 @@ int BPF_PROG(on_sched_fork, struct task_struct *parent,
 }
 
 SEC("tp_btf/sched_process_exec")
-int BPF_PROG(on_sched_exec,struct task_struct *p, u32 old_pid,
+int BPF_PROG(on_sched_exec, struct task_struct *p, u32 old_pid,
 	     struct linux_binprm *prm)
 {
 	struct bpf_event *event;
-	u32 *lctx;
+	u32		 *lctx;
 
 	if (!enable_bpf_events)
 		return 0;
@@ -972,11 +1039,11 @@ int BPF_PROG(on_sched_exec,struct task_struct *p, u32 old_pid,
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = EXEC;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
+	event->type		  = EXEC;
+	event->cpu		  = bpf_get_smp_processor_id();
+	event->ts		  = bpf_ktime_get_ns();
 	event->event.exec.old_pid = old_pid;
-	event->event.exec.pid = BPF_CORE_READ(p, pid);
+	event->event.exec.pid	  = BPF_CORE_READ(p, pid);
 
 	if (layered && (lctx = try_lookup_layered_task_ctx(p)))
 		event->event.exec.layer_id = lctx[LAYER_ID_INDEX];
@@ -991,7 +1058,7 @@ int BPF_PROG(on_sched_exec,struct task_struct *p, u32 old_pid,
 SEC("?tp_btf/sched_process_wait")
 int BPF_PROG(on_sched_wait, struct pid *pid)
 {
-	struct bpf_event *event;
+	struct bpf_event   *event;
 	struct task_struct *p;
 
 	if (!enable_bpf_events || !should_sample())
@@ -1001,18 +1068,17 @@ int BPF_PROG(on_sched_wait, struct pid *pid)
 		return -ENOMEM;
 
 	event->type = WAIT;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
-	p = (struct task_struct *)bpf_get_current_task();
+	event->cpu  = bpf_get_smp_processor_id();
+	event->ts   = bpf_ktime_get_ns();
+	p	    = (struct task_struct *)bpf_get_current_task();
 	if (p) {
 		bpf_core_read_str(&event->event.wait.comm,
-				  sizeof(event->event.wait.comm),
-				  &p->comm);
-		event->event.wait.pid = BPF_CORE_READ(p, pid);
+				  sizeof(event->event.wait.comm), &p->comm);
+		event->event.wait.pid  = BPF_CORE_READ(p, pid);
 		event->event.wait.prio = BPF_CORE_READ(p, prio);
 	} else {
 		__builtin_memset(event->event.wait.comm, 0, MAX_COMM);
-		event->event.wait.pid = 0;
+		event->event.wait.pid  = 0;
 		event->event.wait.prio = 0;
 	}
 
@@ -1032,11 +1098,11 @@ int BPF_PROG(on_gpu_memory_total, u32 gpu, u32 pid, u64 size)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = GPU_MEM;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
-	event->event.gm.gpu = gpu;
-	event->event.gm.pid = pid;
+	event->type	     = GPU_MEM;
+	event->cpu	     = bpf_get_smp_processor_id();
+	event->ts	     = bpf_ktime_get_ns();
+	event->event.gm.gpu  = gpu;
+	event->event.gm.pid  = pid;
 	event->event.gm.size = size;
 
 	bpf_ringbuf_submit(event, 0);
@@ -1047,7 +1113,7 @@ int BPF_PROG(on_gpu_memory_total, u32 gpu, u32 pid, u64 size)
 SEC("tp_btf/cpuhp_enter")
 int BPF_PROG(on_cpuhp_enter, u32 cpu, int target, int state)
 {
-	struct bpf_event *event;
+	struct bpf_event   *event;
 	struct task_struct *p;
 
 	if (!enable_bpf_events || !should_sample())
@@ -1056,13 +1122,13 @@ int BPF_PROG(on_cpuhp_enter, u32 cpu, int target, int state)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = CPU_HP_ENTER;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
-	event->event.chp.cpu = cpu;
+	event->type		= CPU_HP_ENTER;
+	event->cpu		= bpf_get_smp_processor_id();
+	event->ts		= bpf_ktime_get_ns();
+	event->event.chp.cpu	= cpu;
 	event->event.chp.target = target;
-	event->event.chp.state = state;
-	p = (struct task_struct *)bpf_get_current_task();
+	event->event.chp.state	= state;
+	p			= (struct task_struct *)bpf_get_current_task();
 	if (p)
 		event->event.chp.pid = BPF_CORE_READ(p, pid);
 	else
@@ -1076,7 +1142,7 @@ int BPF_PROG(on_cpuhp_enter, u32 cpu, int target, int state)
 SEC("tp_btf/cpuhp_exit")
 int BPF_PROG(on_cpuhp_exit, u32 cpu, int state, int idx, int ret)
 {
-	struct bpf_event *event;
+	struct bpf_event   *event;
 	struct task_struct *p;
 
 	if (!enable_bpf_events || !should_sample())
@@ -1085,14 +1151,14 @@ int BPF_PROG(on_cpuhp_exit, u32 cpu, int state, int idx, int ret)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = CPU_HP_EXIT;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
-	event->event.cxp.cpu = cpu;
+	event->type	       = CPU_HP_EXIT;
+	event->cpu	       = bpf_get_smp_processor_id();
+	event->ts	       = bpf_ktime_get_ns();
+	event->event.cxp.cpu   = cpu;
 	event->event.cxp.state = state;
-	event->event.cxp.idx = idx;
-	event->event.cxp.ret = ret;
-	p = (struct task_struct *)bpf_get_current_task();
+	event->event.cxp.idx   = idx;
+	event->event.cxp.ret   = ret;
+	p		       = (struct task_struct *)bpf_get_current_task();
 	if (p)
 		event->event.cxp.pid = BPF_CORE_READ(p, pid);
 	else
@@ -1114,11 +1180,11 @@ int BPF_PROG(on_hw_pressure_update, u32 cpu, u64 hw_pressure)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = HW_PRESSURE;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
+	event->type		     = HW_PRESSURE;
+	event->cpu		     = bpf_get_smp_processor_id();
+	event->ts		     = bpf_ktime_get_ns();
 	event->event.hwp.hw_pressure = hw_pressure;
-	event->event.hwp.cpu = cpu;
+	event->event.hwp.cpu	     = cpu;
 
 	bpf_ringbuf_submit(event, 0);
 
@@ -1128,10 +1194,10 @@ int BPF_PROG(on_hw_pressure_update, u32 cpu, u64 hw_pressure)
 SEC("perf_event")
 int perf_sample_handler(struct bpf_perf_event_data *ctx)
 {
-	struct bpf_event *event;
+	struct bpf_event   *event;
 	struct task_struct *task;
-	u32 *lctx;
-	int ret;
+	u32		   *lctx;
+	int		    ret;
 
 	if (!enable_bpf_events || !should_sample())
 		return 0;
@@ -1139,11 +1205,11 @@ int perf_sample_handler(struct bpf_perf_event_data *ctx)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = PERF_SAMPLE;
-	event->cpu = bpf_get_smp_processor_id();
-	event->ts = bpf_ktime_get_ns();
+	event->type		     = PERF_SAMPLE;
+	event->cpu		     = bpf_get_smp_processor_id();
+	event->ts		     = bpf_ktime_get_ns();
 	event->event.perf_sample.pid = bpf_get_current_pid_tgid() & 0xffffffff;
-	event->event.perf_sample.instruction_pointer = ctx->regs.ip;
+	event->event.perf_sample.instruction_pointer = PT_REGS_IP(&ctx->regs);
 	event->event.perf_sample.cpu_id = bpf_get_smp_processor_id();
 
 	// Get current task for layer ID lookup
@@ -1165,14 +1231,14 @@ int perf_sample_handler(struct bpf_perf_event_data *ctx)
 		event->event.perf_sample.kernel_stack_size = 0;
 		// Try fallback method for kernel stack - get current IP
 		if (event->event.perf_sample.is_kernel) {
-			event->event.perf_sample.kernel_stack[0] = ctx->regs.ip;
+			event->event.perf_sample.kernel_stack[0] =
+				PT_REGS_IP(&ctx->regs);
 			event->event.perf_sample.kernel_stack_size = 1;
 		}
 	}
 
 	// Capture user stack trace
-	ret = bpf_get_stack(ctx,
-			    event->event.perf_sample.user_stack,
+	ret = bpf_get_stack(ctx, event->event.perf_sample.user_stack,
 			    sizeof(event->event.perf_sample.user_stack),
 			    BPF_F_USER_STACK); // Only user stack flag
 	if (ret > 0) {
@@ -1181,11 +1247,13 @@ int perf_sample_handler(struct bpf_perf_event_data *ctx)
 		event->event.perf_sample.user_stack_size = 0;
 		// Try fallback method for user stack - get current IP if userspace
 		if (!event->event.perf_sample.is_kernel) {
-			event->event.perf_sample.user_stack[0] = ctx->regs.ip;
+			event->event.perf_sample.user_stack[0] =
+				PT_REGS_IP(&ctx->regs);
 			event->event.perf_sample.user_stack_size = 1;
 		}
 	}
-	event->event.perf_sample.is_kernel = event->event.perf_sample.user_stack_size <= 1;
+	event->event.perf_sample.is_kernel =
+		event->event.perf_sample.user_stack_size <= 1;
 
 	bpf_ringbuf_submit(event, 0);
 
@@ -1196,9 +1264,10 @@ SEC("syscall")
 int BPF_PROG(scxtop_init)
 {
 	struct timer_wrapper *timerw;
-	int timer_id, err;
+	int		      timer_id, err;
 
-	bpf_for(timer_id, 0, MAX_TIMERS) {
+	bpf_for(timer_id, 0, MAX_TIMERS)
+	{
 		timerw = bpf_map_lookup_elem(&timers, &timer_id);
 		if (!timerw)
 			return 0;
@@ -1209,7 +1278,8 @@ int BPF_PROG(scxtop_init)
 		if (err)
 			return 0;
 
-		err = bpf_timer_set_callback(&timerw->timer, &stop_trace_timer_callback);
+		err = bpf_timer_set_callback(&timerw->timer,
+					     &stop_trace_timer_callback);
 		if (err)
 			return 0;
 	}
