@@ -45,6 +45,7 @@ impl Default for KeyMap {
         bindings.insert(Key::Char('T'), Action::SetState(AppState::PerfTop));
         bindings.insert(Key::Char('f'), Action::Filter);
         bindings.insert(Key::Char('u'), Action::ToggleUncoreFreq);
+        bindings.insert(Key::Char('F'), Action::ToggleCpuFreq);
         bindings.insert(Key::Char('L'), Action::ToggleLocalization);
         bindings.insert(Key::Char('P'), Action::ToggleHwPressure);
         bindings.insert(Key::Char('h'), Action::SetState(AppState::Help));
@@ -62,6 +63,7 @@ impl Default for KeyMap {
         bindings.insert(Key::Char('j'), Action::PrevEvent);
         bindings.insert(Key::Char('k'), Action::NextEvent);
         bindings.insert(Key::Char('q'), Action::Quit);
+        bindings.insert(Key::Char('Q'), Action::Quit);
         bindings.insert(Key::Char('t'), Action::ChangeTheme);
         bindings.insert(Key::Char('-'), Action::DecTickRate);
         bindings.insert(Key::Char('+'), Action::IncTickRate);
