@@ -119,6 +119,7 @@ pub enum AppState {
 pub enum ViewState {
     Sparkline,
     BarChart,
+    LineGauge,
 }
 
 impl ViewState {
@@ -126,7 +127,8 @@ impl ViewState {
     pub fn next(&self) -> Self {
         match self {
             ViewState::Sparkline => ViewState::BarChart,
-            ViewState::BarChart => ViewState::Sparkline,
+            ViewState::BarChart => ViewState::LineGauge,
+            ViewState::LineGauge => ViewState::Sparkline,
         }
     }
 }
@@ -136,6 +138,7 @@ impl std::fmt::Display for ViewState {
         match self {
             ViewState::Sparkline => write!(f, "sparkline"),
             ViewState::BarChart => write!(f, "barchart"),
+            ViewState::LineGauge => write!(f, "linegauge"),
         }
     }
 }
@@ -291,6 +294,8 @@ pub struct SchedWakeActionCtx {
     pub tgid: u32,
     pub prio: i32,
     pub comm: SsoString,
+    pub waker_pid: u32,
+    pub waker_comm: SsoString,
 }
 
 pub type SchedWakeupNewAction = SchedWakeActionCtx;
@@ -571,6 +576,7 @@ impl TryFrom<&bpf_event> for Action {
                 let wakeup = unsafe { event.event.wakeup };
 
                 let comm = String::from_utf8_lossy(&wakeup.comm);
+                let waker_comm = String::from_utf8_lossy(&wakeup.waker_comm);
 
                 Ok(Action::SchedWakeup(SchedWakeupAction {
                     ts: event.ts,
@@ -579,6 +585,8 @@ impl TryFrom<&bpf_event> for Action {
                     tgid: wakeup.tgid,
                     prio: wakeup.prio,
                     comm: comm.into(),
+                    waker_pid: wakeup.waker_pid,
+                    waker_comm: waker_comm.into(),
                 }))
             }
             #[allow(non_upper_case_globals)]
@@ -586,6 +594,7 @@ impl TryFrom<&bpf_event> for Action {
                 let waking = unsafe { &event.event.waking };
 
                 let comm = String::from_utf8_lossy(&waking.comm);
+                let waker_comm = String::from_utf8_lossy(&waking.waker_comm);
 
                 Ok(Action::SchedWaking(SchedWakingAction {
                     ts: event.ts,
@@ -594,6 +603,8 @@ impl TryFrom<&bpf_event> for Action {
                     tgid: waking.tgid,
                     prio: waking.prio,
                     comm: comm.into(),
+                    waker_pid: waking.waker_pid,
+                    waker_comm: waker_comm.into(),
                 }))
             }
             #[allow(non_upper_case_globals)]
