@@ -4,6 +4,7 @@
 // GNU General Public License version 2.
 
 use anyhow::Result;
+use procfs::process::Process as ProcProcess;
 use rand::rngs::StdRng;
 use rand::RngCore;
 use rand::SeedableRng;
@@ -14,10 +15,11 @@ use std::fs;
 use std::time::{SystemTime, UNIX_EPOCH};
 
 use crate::edm::ActionHandler;
+use crate::util::get_clock_value;
 use crate::{
     Action, CpuhpEnterAction, CpuhpExitAction, ExecAction, ExitAction, ForkAction, GpuMemAction,
-    IPIAction, KprobeAction, SchedMigrateTaskAction, SchedSwitchAction, SchedWakeupAction,
-    SchedWakingAction, SoftIRQAction,
+    IPIAction, KprobeAction, SchedHangAction, SchedMigrateTaskAction, SchedSwitchAction,
+    SchedWakeupAction, SchedWakingAction, SoftIRQAction, SystemStatAction, WaitAction,
 };
 
 use perfetto_protos::{
@@ -32,18 +34,21 @@ use perfetto_protos::{
     ipi::IpiRaiseFtraceEvent,
     irq::{SoftirqEntryFtraceEvent, SoftirqExitFtraceEvent},
     process_descriptor::ProcessDescriptor,
+    process_tree::{process_tree::Process, ProcessTree},
     sched::{
         SchedMigrateTaskFtraceEvent, SchedProcessExecFtraceEvent, SchedProcessExitFtraceEvent,
-        SchedProcessForkFtraceEvent, SchedSwitchFtraceEvent, SchedWakeupFtraceEvent,
-        SchedWakingFtraceEvent,
+        SchedProcessForkFtraceEvent, SchedProcessHangFtraceEvent, SchedProcessWaitFtraceEvent,
+        SchedSwitchFtraceEvent, SchedWakeupFtraceEvent, SchedWakingFtraceEvent,
     },
+    sys_stats::{sys_stats::CpuTimes, sys_stats::MeminfoValue, SysStats},
+    sys_stats_counters::MeminfoCounters,
     thread_descriptor::ThreadDescriptor,
     trace::Trace,
     trace_packet::{trace_packet, TracePacket},
     track_descriptor::{track_descriptor::Static_or_dynamic_name, TrackDescriptor},
     track_event::{track_event, TrackEvent},
 };
-use protobuf::{Message, SpecialFields};
+use protobuf::{EnumOrUnknown, Message, SpecialFields};
 
 /// Handler for perfetto traces. For details on data flow in perfetto see:
 /// https://perfetto.dev/docs/concepts/buffers and
@@ -60,13 +65,15 @@ pub struct PerfettoTraceManager {
     // per cpu ftrace events
     ftrace_events: BTreeMap<u32, Vec<FtraceEvent>>,
     dsq_lat_events: BTreeMap<u64, Vec<TrackEvent>>,
-    dsq_lat_trusted_packet_seq_uuid: u32,
     dsq_nr_queued_events: BTreeMap<u64, Vec<TrackEvent>>,
-    dsq_nr_queued_trusted_packet_seq_uuid: u32,
     dsq_uuids: BTreeMap<u64, u64>,
-    processes: HashMap<u64, ProcessDescriptor>,
+    process_descriptors: HashMap<u64, ProcessDescriptor>,
+    processes: HashMap<u64, Process>,
     threads: HashMap<u64, ThreadDescriptor>,
     process_uuids: HashMap<i32, u64>,
+    sys_stats: BTreeMap<u64, Vec<SysStats>>,
+    mem_events: BTreeMap<String, Vec<TrackEvent>>,
+    mem_uuids: HashMap<String, u64>,
 }
 
 impl PerfettoTraceManager {
@@ -78,13 +85,14 @@ impl PerfettoTraceManager {
                 .expect("Time went backwards")
                 .as_secs(),
         );
+
         let mut rng = StdRng::seed_from_u64(trace_uuid);
-        let trace = Trace::default();
-        let dsq_lat_trusted_packet_seq_uuid = rng.next_u32();
-        let dsq_nr_queued_trusted_packet_seq_uuid = rng.next_u32();
+        let mut mem_uuids = HashMap::new();
+        mem_uuids.insert("mem_ratio".to_string(), rng.next_u64());
+        mem_uuids.insert("swap_ratio".to_string(), rng.next_u64());
 
         Self {
-            trace,
+            trace: Trace::default(),
             trace_id: 0,
             trusted_pid: std::process::id() as i32,
             rng,
@@ -92,12 +100,14 @@ impl PerfettoTraceManager {
             ftrace_events: BTreeMap::new(),
             dsq_uuids: BTreeMap::new(),
             dsq_lat_events: BTreeMap::new(),
-            dsq_lat_trusted_packet_seq_uuid,
             dsq_nr_queued_events: BTreeMap::new(),
-            dsq_nr_queued_trusted_packet_seq_uuid,
+            process_descriptors: HashMap::new(),
             processes: HashMap::new(),
             threads: HashMap::new(),
             process_uuids: HashMap::new(),
+            sys_stats: BTreeMap::new(),
+            mem_events: BTreeMap::new(),
+            mem_uuids,
         }
     }
 
@@ -164,15 +174,27 @@ impl PerfettoTraceManager {
             desc_map.insert(dsq_uuid, descs);
         }
 
-        desc_map
-    }
-
-    fn get_clock_value(&mut self, clock_id: libc::c_int) -> u64 {
-        let mut ts: libc::timespec = unsafe { std::mem::zeroed() };
-        if unsafe { libc::clock_gettime(clock_id, &mut ts) } != 0 {
-            return 0;
+        for (name, &uuid) in &self.mem_uuids {
+            desc_map.insert(
+                uuid,
+                vec![TrackDescriptor {
+                    uuid: Some(uuid),
+                    counter: Some(CounterDescriptor {
+                        unit: Some(UNIT_COUNT.into()),
+                        unit_name: Some(name.to_string()),
+                        is_incremental: Some(false),
+                        ..CounterDescriptor::default()
+                    })
+                    .into(),
+                    static_or_dynamic_name: Some(Static_or_dynamic_name::StaticName(
+                        name.to_string(),
+                    )),
+                    ..TrackDescriptor::default()
+                }],
+            );
         }
-        (ts.tv_sec as u64 * 1_000_000_000) + ts.tv_nsec as u64
+
+        desc_map
     }
 
     fn snapshot_clocks(&mut self) {
@@ -180,32 +202,32 @@ impl PerfettoTraceManager {
             clocks: vec![
                 Clock {
                     clock_id: Some(BuiltinClock::BUILTIN_CLOCK_MONOTONIC as u32),
-                    timestamp: Some(self.get_clock_value(libc::CLOCK_MONOTONIC)),
+                    timestamp: Some(get_clock_value(libc::CLOCK_MONOTONIC)),
                     ..Clock::default()
                 },
                 Clock {
                     clock_id: Some(BuiltinClock::BUILTIN_CLOCK_BOOTTIME as u32),
-                    timestamp: Some(self.get_clock_value(libc::CLOCK_BOOTTIME)),
+                    timestamp: Some(get_clock_value(libc::CLOCK_BOOTTIME)),
                     ..Clock::default()
                 },
                 Clock {
                     clock_id: Some(BuiltinClock::BUILTIN_CLOCK_REALTIME as u32),
-                    timestamp: Some(self.get_clock_value(libc::CLOCK_REALTIME)),
+                    timestamp: Some(get_clock_value(libc::CLOCK_REALTIME)),
                     ..Clock::default()
                 },
                 Clock {
                     clock_id: Some(BuiltinClock::BUILTIN_CLOCK_REALTIME_COARSE as u32),
-                    timestamp: Some(self.get_clock_value(libc::CLOCK_REALTIME_COARSE)),
+                    timestamp: Some(get_clock_value(libc::CLOCK_REALTIME_COARSE)),
                     ..Clock::default()
                 },
                 Clock {
                     clock_id: Some(BuiltinClock::BUILTIN_CLOCK_MONOTONIC_COARSE as u32),
-                    timestamp: Some(self.get_clock_value(libc::CLOCK_MONOTONIC_COARSE)),
+                    timestamp: Some(get_clock_value(libc::CLOCK_MONOTONIC_COARSE)),
                     ..Clock::default()
                 },
                 Clock {
                     clock_id: Some(BuiltinClock::BUILTIN_CLOCK_MONOTONIC_RAW as u32),
-                    timestamp: Some(self.get_clock_value(libc::CLOCK_MONOTONIC_RAW)),
+                    timestamp: Some(get_clock_value(libc::CLOCK_MONOTONIC_RAW)),
                     ..Clock::default()
                 },
             ],
@@ -219,48 +241,68 @@ impl PerfettoTraceManager {
         });
     }
 
-    fn generate_key(&mut self, v1: u32, v2: u32) -> u64 {
+    fn generate_key(&self, v1: u32, v2: u32) -> u64 {
         let v1_u32 = v1 as u64;
         let v2_u32 = v2 as u64;
         (v1_u32 << 32) | v2_u32
     }
 
     fn record_process_thread(&mut self, pid: u32, tid: u32, comm: String) {
-        let key = self.generate_key(pid, tid);
+        if pid != tid {
+            let thread_key = self.generate_key(pid, tid);
+            self.threads
+                .entry(thread_key)
+                .or_insert_with(|| ThreadDescriptor {
+                    tid: Some(tid as i32),
+                    pid: Some(pid as i32),
+                    thread_name: Some(comm.clone()),
+                    ..ThreadDescriptor::default()
+                });
+        }
 
-        if pid == tid {
-            let process = self
-                .processes
-                .entry(key)
-                .or_insert_with(|| ProcessDescriptor {
+        // Let's check if this is the first time we've seen this process.
+        let parent_key = self.generate_key(pid, pid);
+        if !self.process_descriptors.contains_key(&parent_key) {
+            let process_name = if pid == tid {
+                Some(comm)
+            } else {
+                self.get_comm(pid)
+            };
+            let cmdline = self.get_cmdline(pid);
+            self.process_descriptors.insert(
+                parent_key,
+                ProcessDescriptor {
                     pid: Some(pid as i32),
+                    cmdline: cmdline.clone(),
+                    process_name,
                     ..ProcessDescriptor::default()
-                });
-            process.process_name.get_or_insert(comm);
-        } else {
-            self.threads.entry(key).or_insert_with(|| ThreadDescriptor {
-                tid: Some(tid as i32),
-                pid: Some(pid as i32),
-                thread_name: Some(comm),
-                ..ThreadDescriptor::default()
-            });
-            // Create a ProcessDescriptor with an empty comm if one doesn't
-            // exist - if we ever see the main thread we populate the process
-            // name field there (see above).
-            let pkey = self.generate_key(pid, pid);
-            self.processes
-                .entry(pkey)
-                .or_insert_with(|| ProcessDescriptor {
+                },
+            );
+
+            self.processes.insert(
+                parent_key,
+                Process {
                     pid: Some(pid as i32),
-                    ..ProcessDescriptor::default()
-                });
+                    cmdline,
+                    ..Process::default()
+                },
+            );
         }
     }
 
-    fn get_comm(pid: i32) -> Result<String> {
-        let path = format!("/proc/{}/comm", pid);
-        let content = fs::read_to_string(path)?;
-        Ok(content.trim_end().to_string())
+    fn get_comm(&self, pid: u32) -> Option<String> {
+        let pid = pid.try_into().expect("u32 was not able to fit into i32");
+        let process = ProcProcess::new(pid).ok()?;
+        let stat = process.stat().ok()?;
+        Some(stat.comm)
+    }
+
+    fn get_cmdline(&self, pid: u32) -> Vec<String> {
+        let pid = pid.try_into().expect("u32 was not able to fit into i32");
+        ProcProcess::new(pid)
+            .ok()
+            .and_then(|p| p.cmdline().ok())
+            .unwrap_or_default()
     }
 
     /// Stops the trace and writes to configured output file.
@@ -276,6 +318,7 @@ impl PerfettoTraceManager {
 
         let trace_cpus: Vec<u32> = self.ftrace_events.keys().cloned().collect();
         let trace_dsqs: Vec<u64> = self.dsq_nr_queued_events.keys().cloned().collect();
+        let stat_ts: Vec<u64> = self.sys_stats.keys().cloned().collect();
 
         fn timestamp_absolute_us(e: &TrackEvent) -> i64 {
             use perfetto_protos::track_event::track_event::Timestamp;
@@ -301,14 +344,10 @@ impl PerfettoTraceManager {
                 .for_each(|(_, v)| v.retain(|e| e.timestamp.unwrap_or(0) < ns));
         };
 
-        for (_, mut process) in self.processes.drain() {
+        for (_, process) in self.process_descriptors.drain() {
             let uuid = self.rng.next_u64();
             self.process_uuids.insert(process.pid(), uuid);
 
-            if process.process_name().is_empty() {
-                process.process_name = Self::get_comm(process.pid()).ok();
-            }
-
             let desc = TrackDescriptor {
                 uuid: Some(uuid),
                 process: Some(process).into(),
@@ -322,6 +361,19 @@ impl PerfettoTraceManager {
             self.trace.packet.push(packet);
         }
 
+        for (_, process) in self.processes.drain() {
+            let tree = ProcessTree {
+                processes: vec![process],
+                ..ProcessTree::default()
+            };
+
+            let packet = TracePacket {
+                data: Some(trace_packet::Data::ProcessTree(tree)),
+                ..TracePacket::default()
+            };
+            self.trace.packet.push(packet);
+        }
+
         for (_, thread) in self.threads.drain() {
             let uuid = self.rng.next_u64();
 
@@ -350,6 +402,7 @@ impl PerfettoTraceManager {
         }
 
         // dsq latency tracks
+        let dsq_lat_trusted_packet_seq_uuid = self.rng.next_u32();
         for dsq in &trace_dsqs {
             if let Some(events) = self.dsq_lat_events.remove(dsq) {
                 for dsq_lat_event in events {
@@ -359,7 +412,7 @@ impl PerfettoTraceManager {
                         timestamp: Some(ts),
                         optional_trusted_packet_sequence_id: Some(
                             trace_packet::Optional_trusted_packet_sequence_id::TrustedPacketSequenceId(
-                                self.dsq_lat_trusted_packet_seq_uuid,
+                                dsq_lat_trusted_packet_seq_uuid,
                             ),
                         ),
                         ..TracePacket::default()
@@ -369,6 +422,7 @@ impl PerfettoTraceManager {
         }
 
         // dsq nr_queued tracks
+        let dsq_nr_queued_trusted_packet_seq_uuid = self.rng.next_u32();
         for dsq in &trace_dsqs {
             if let Some(events) = self.dsq_nr_queued_events.remove(dsq) {
                 for dsq_nr_queued_event in events {
@@ -378,7 +432,7 @@ impl PerfettoTraceManager {
                         timestamp: Some(ts),
                         optional_trusted_packet_sequence_id: Some(
                             trace_packet::Optional_trusted_packet_sequence_id::TrustedPacketSequenceId(
-                                self.dsq_nr_queued_trusted_packet_seq_uuid,
+                                dsq_nr_queued_trusted_packet_seq_uuid,
                             ),
                         ),
                         ..TracePacket::default()
@@ -387,6 +441,37 @@ impl PerfettoTraceManager {
             }
         }
 
+        // system stat events
+        for ts in &stat_ts {
+            if let Some(events) = self.sys_stats.remove(ts) {
+                for sys_stat_event in events {
+                    self.trace.packet.push(TracePacket {
+                        data: Some(trace_packet::Data::SysStats(sys_stat_event)),
+                        timestamp: Some(*ts),
+                        ..TracePacket::default()
+                    });
+                }
+            }
+        }
+
+        // mem events
+        for events in self.mem_events.values_mut() {
+            let mem_sequence_id = self.rng.next_u32();
+            for mem_event in events.drain(..) {
+                let ts: u64 = timestamp_absolute_us(&mem_event) as u64 / 1_000;
+                self.trace.packet.push(TracePacket {
+                    data: Some(trace_packet::Data::TrackEvent(mem_event)),
+                    timestamp: Some(ts),
+                    optional_trusted_packet_sequence_id: Some(
+                        trace_packet::Optional_trusted_packet_sequence_id::TrustedPacketSequenceId(
+                            mem_sequence_id,
+                        ),
+                    ),
+                    ..TracePacket::default()
+                });
+            }
+        }
+
         // ftrace events
         for cpu in &trace_cpus {
             self.trace.packet.push(TracePacket {
@@ -460,6 +545,7 @@ impl PerfettoTraceManager {
             child_pid,
             parent_comm,
             child_comm,
+            ..
         } = action;
 
         self.ftrace_events.entry(*cpu).or_default().push({
@@ -486,6 +572,7 @@ impl PerfettoTraceManager {
             cpu,
             old_pid,
             pid,
+            ..
         } = action;
 
         self.ftrace_events.entry(*cpu).or_default().push({
@@ -504,6 +591,32 @@ impl PerfettoTraceManager {
         });
     }
 
+    pub fn on_wait(&mut self, action: &WaitAction) {
+        let WaitAction {
+            ts,
+            cpu,
+            comm,
+            pid,
+            prio,
+        } = action;
+
+        self.ftrace_events.entry(*cpu).or_default().push({
+            FtraceEvent {
+                timestamp: Some(*ts),
+                pid: Some(*pid),
+                event: Some(ftrace_event::Event::SchedProcessWait(
+                    SchedProcessWaitFtraceEvent {
+                        comm: Some(comm.as_str().to_string()),
+                        pid: Some((*pid).try_into().unwrap()),
+                        prio: Some(*prio),
+                        special_fields: SpecialFields::new(),
+                    },
+                )),
+                ..FtraceEvent::default()
+            }
+        });
+    }
+
     /// Adds events for on sched_wakeup.
     pub fn on_sched_wakeup(&mut self, action: &SchedWakeupAction) {
         let SchedWakeupAction {
@@ -513,6 +626,7 @@ impl PerfettoTraceManager {
             tgid,
             prio,
             comm,
+            ..
         } = action;
 
         self.ftrace_events.entry(*cpu).or_default().push({
@@ -546,6 +660,7 @@ impl PerfettoTraceManager {
             tgid,
             prio,
             comm,
+            ..
         } = action;
 
         self.ftrace_events.entry(*cpu).or_default().push({
@@ -574,6 +689,7 @@ impl PerfettoTraceManager {
             pid,
             prio,
             comm,
+            ..
         } = action;
 
         self.ftrace_events.entry(*cpu).or_default().push({
@@ -594,6 +710,26 @@ impl PerfettoTraceManager {
         });
     }
 
+    /// Adds events for on sched_hang.
+    pub fn on_sched_hang(&mut self, action: &SchedHangAction) {
+        let SchedHangAction { ts, cpu, comm, pid } = action;
+
+        self.ftrace_events.entry(*cpu).or_default().push({
+            FtraceEvent {
+                timestamp: Some(*ts),
+                pid: Some(*pid),
+                event: Some(ftrace_event::Event::SchedProcessHang(
+                    SchedProcessHangFtraceEvent {
+                        comm: Some(comm.as_str().to_string()),
+                        pid: Some((*pid).try_into().unwrap()),
+                        special_fields: SpecialFields::new(),
+                    },
+                )),
+                ..FtraceEvent::default()
+            }
+        });
+    }
+
     /// Adds events for the softirq entry/exit events.
     pub fn on_softirq(&mut self, action: &SoftIRQAction) {
         self.ftrace_events.entry(action.cpu).or_default().extend({
@@ -741,6 +877,113 @@ impl PerfettoTraceManager {
         });
     }
 
+    fn meminfo_value(key: MeminfoCounters, value: u64) -> MeminfoValue {
+        MeminfoValue {
+            key: Some(EnumOrUnknown::new(key)),
+            value: Some(value),
+            special_fields: SpecialFields::default(),
+        }
+    }
+
+    pub fn on_sys_stat(&mut self, action: &SystemStatAction) {
+        let SystemStatAction {
+            ts,
+            cpu_data_prev,
+            cpu_data_current,
+            mem_info,
+        } = action;
+
+        let mut cpufreq_khz = Vec::with_capacity(cpu_data_prev.len());
+        let mut cpu_stat = Vec::with_capacity(cpu_data_prev.len());
+        for (cpu, data) in cpu_data_current.iter() {
+            cpufreq_khz.insert(
+                *cpu,
+                (data.freq_khz * 1000)
+                    .try_into()
+                    .expect("Should have been able to convert u64 to u32"),
+            );
+
+            // Skip if we haven't collected prev data yet for this cpu
+            if !cpu_data_prev.contains_key(cpu) {
+                continue;
+            }
+
+            let current_data = &data.cpu_util_data;
+            let prev_data = &cpu_data_prev
+                .get(cpu)
+                .expect("Should have cpu data")
+                .cpu_util_data;
+
+            let cpu_time = CpuTimes {
+                cpu_id: Some(*cpu as u32),
+                user_ns: Some(current_data.user - prev_data.user),
+                user_nice_ns: Some(current_data.nice - prev_data.nice),
+                system_mode_ns: Some(current_data.system - prev_data.system),
+                idle_ns: Some(current_data.idle - prev_data.idle),
+                io_wait_ns: Some(current_data.iowait - prev_data.iowait),
+                irq_ns: Some(current_data.irq - prev_data.irq),
+                softirq_ns: Some(current_data.softirq - prev_data.softirq),
+                special_fields: SpecialFields::new(),
+                steal_ns: None,
+            };
+            cpu_stat.insert(*cpu, cpu_time);
+        }
+
+        let mem_data = vec![
+            Self::meminfo_value(MeminfoCounters::MEMINFO_MEM_FREE, mem_info.free_kb),
+            Self::meminfo_value(MeminfoCounters::MEMINFO_SWAP_FREE, mem_info.swap_free_kb),
+        ];
+
+        self.sys_stats.entry(*ts).or_default().push({
+            SysStats {
+                cpu_stat,
+                cpufreq_khz,
+                meminfo: mem_data,
+                ..SysStats::default()
+            }
+        });
+
+        self.mem_events
+            .entry("mem_ratio".to_string())
+            .or_default()
+            .push(TrackEvent {
+                type_: Some(track_event::Type::TYPE_COUNTER.into()),
+                track_uuid: Some(
+                    *self
+                        .mem_uuids
+                        .get("mem_ratio")
+                        .expect("Should have mem_ratio"),
+                ),
+                counter_value_field: Some(track_event::Counter_value_field::DoubleCounterValue(
+                    mem_info.free_ratio() * 100.0,
+                )),
+                timestamp: Some(track_event::Timestamp::TimestampAbsoluteUs(
+                    (*ts) as i64 / 1000,
+                )),
+                ..TrackEvent::default()
+            });
+
+        self.mem_events
+            .entry("swap_ratio".to_string())
+            .or_default()
+            .push(TrackEvent {
+                type_: Some(track_event::Type::TYPE_COUNTER.into()),
+                track_uuid: Some(
+                    *self
+                        .mem_uuids
+                        .get("swap_ratio")
+                        .expect("Should have swap_ratio"),
+                ),
+                counter_value_field: Some(track_event::Counter_value_field::DoubleCounterValue(
+                    mem_info.swap_ratio() * 100.0,
+                )),
+                timestamp: Some(track_event::Timestamp::TimestampAbsoluteUs(
+                    (*ts) as i64 / 1000,
+                )),
+                ..TrackEvent::default()
+            });
+    }
+
     /// Adds events for the sched_switch event.
     pub fn on_sched_switch(&mut self, action: &SchedSwitchAction) {
         let SchedSwitchAction {
@@ -871,6 +1114,9 @@ impl ActionHandler for PerfettoTraceManager {
             Action::Kprobe(a) => {
                 self.on_kprobe(a);
             }
+            Action::SystemStat(a) => {
+                self.on_sys_stat(a);
+            }
             _ => {}
         }
 
